# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    BATCHED,
)
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    BOUNDING_BOXES,
)
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    IMAGES,
)
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    LABELS,
)
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    SEGMENTATION_MASKS,
)
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils


@keras_cv_export("keras_cv.layers.Mosaic")
class Mosaic(VectorizedBaseImageAugmentationLayer):
    """Mosaic implements the mosaic data augmentation technique.

    Mosaic data augmentation first takes 4 images from the batch and makes a
    grid. After that based on the offset, a crop is taken to form the mosaic
    image. Labels are in the same ratio as the area of their images in the
    output image. Bounding boxes are translated according to the position of the
    4 images.

    Args:
        offset: A tuple of two floats, a single float or
            `keras_cv.FactorSampler`. `offset` is used to determine the offset
            of the mosaic center from the top-left corner of the mosaic. If a
            tuple is used, the x and y coordinates of the mosaic center are
            sampled between the two values for every image augmented. If a
            single float is used, a value between `0.0` and the passed float is
            sampled. In order to ensure the value is always the same, please
            pass a tuple with two identical floats: `(0.5, 0.5)`. Defaults to
            (0.25, 0.75).
        bounding_box_format: a case-insensitive string (for example, "xyxy") to
            be passed if bounding boxes are being augmented by this layer. Each
            bounding box is defined by at least these 4 values. The inputs may
            contain additional information such as classes and confidence after
            these 4 values but these values will be ignored and returned as is.
            For detailed information on the supported formats, see the
            [KerasCV bounding box documentation](https://keras.io/api/keras_cv/bounding_box/formats/).
            Defaults to None.
        seed: integer, used to create a random seed.

    References:
        - [Yolov4 paper](https://arxiv.org/pdf/2004.10934).
        - [Yolov5 implementation](https://github.com/ultralytics/yolov5).
        - [YoloX implementation](https://github.com/Megvii-BaseDetection/YOLOX)

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    labels = tf.one_hot(labels,10)
    labels = tf.cast(tf.squeeze(labels), tf.float32)
    mosaic = keras_cv.layers.preprocessing.Mosaic()
    output = mosaic({'images': images, 'labels': labels})
    # output == {'images': updated_images, 'labels': updated_labels}
    ```
    """  # noqa: E501

    def __init__(
        self, offset=(0.25, 0.75), bounding_box_format=None, seed=None, **kwargs
    ):
        super().__init__(seed=seed, **kwargs)
        self.offset = offset
        self.bounding_box_format = bounding_box_format
        self.center_sampler = preprocessing_utils.parse_factor(
            offset, param_name="offset", seed=seed
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # pick 3 indices for every batch to create the mosaic output with.
        permutation_order = self._random_generator.uniform(
            (batch_size, 3),
            minval=0,
            maxval=batch_size,
            dtype=tf.int32,
        )
        # concatenate the batches with permutation order to get all 4 images of
        # the mosaic
        permutation_order = tf.concat(
            [tf.expand_dims(tf.range(batch_size), axis=-1), permutation_order],
            axis=-1,
        )

        mosaic_centers_x = self.center_sampler(
            shape=(batch_size,), dtype=self.compute_dtype
        )
        mosaic_centers_y = self.center_sampler(
            shape=(batch_size,), dtype=self.compute_dtype
        )
        mosaic_centers = tf.stack((mosaic_centers_x, mosaic_centers_y), axis=-1)

        return {
            "permutation_order": permutation_order,
            "mosaic_centers": mosaic_centers,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        raise ValueError(
            "Mosaic received ragged images to `call`. The layer relies on "
            "combining multiple examples with same size, and as such will not "
            "behave as expected. Please call the layer with dense images with "
            "same size. This is an implementation constraint, not an algorithm "
            "constraint. If you find this method helpful, please open an issue "
            "on KerasCV."
        )

    def augment_images(
        self, images, transformations, resize_method="bilinear", **kwargs
    ):
        batch_size = tf.shape(images)[0]
        input_height, input_width, _ = images.shape[1:]

        # forms mosaic for one image from the batch
        permutation_order = transformations["permutation_order"]
        mosaic_images = tf.gather(images, permutation_order)

        tops = tf.concat([mosaic_images[:, 0], mosaic_images[:, 1]], axis=2)
        bottoms = tf.concat([mosaic_images[:, 2], mosaic_images[:, 3]], axis=2)
        outputs = tf.concat([tops, bottoms], axis=1)

        # cropping coordinates for the mosaic
        mosaic_centers = transformations["mosaic_centers"]
        mosaic_centers_x = mosaic_centers[..., 0] * input_width
        mosaic_centers_y = mosaic_centers[..., 1] * input_height
        x1s = (input_width - mosaic_centers_x) / (input_width * 2 - 1)
        y1s = (input_height - mosaic_centers_y) / (input_height * 2 - 1)
        x2s = x1s + (input_width) / (input_width * 2 - 1)
        y2s = y1s + (input_height) / (input_height * 2 - 1)
        cropping_boxes = tf.stack([y1s, x1s, y2s, x2s], axis=-1)

        # helps avoid retracing caused by slicing, inspired by RRC
        # implementation
        # boxes must be type tf.float32
        outputs = tf.image.crop_and_resize(
            outputs,
            tf.cast(cropping_boxes, tf.float32),
            tf.range(batch_size),
            [input_height, input_width],
            method=resize_method,
        )
        # tf.image.crop_and_resize will always output float32, so we need to
        # recast tf.image.crop_and_resize outputs
        # [num_boxes, crop_height, crop_width, depth] since num_boxes is always
        # one we squeeze axis 0
        outputs = tf.cast(outputs, self.compute_dtype)
        return outputs

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return self.augment_images(
            segmentation_masks, transformations, resize_method="nearest"
        )

    def augment_labels(self, labels, transformations, images=None, **kwargs):
        input_height, input_width, _ = images.shape[1:]
        # updates labels for one output mosaic
        permutation_order = transformations["permutation_order"]
        labels_for_mosaic = tf.gather(labels, permutation_order)

        mosaic_centers = transformations["mosaic_centers"]
        center_x = mosaic_centers[..., 0] * input_width
        center_y = mosaic_centers[..., 1] * input_height

        area = input_height * input_width

        # labels are in the same ratio as the area of the images
        top_left_ratio = (center_x * center_y) / area
        top_right_ratio = ((input_width - center_x) * center_y) / area
        bottom_left_ratio = (center_x * (input_height - center_y)) / area
        bottom_right_ratio = (
            (input_width - center_x) * (input_height - center_y)
        ) / area
        labels = (
            labels_for_mosaic[:, 0] * top_left_ratio[:, tf.newaxis]
            + labels_for_mosaic[:, 1] * top_right_ratio[:, tf.newaxis]
            + labels_for_mosaic[:, 2] * bottom_left_ratio[:, tf.newaxis]
            + labels_for_mosaic[:, 3] * bottom_right_ratio[:, tf.newaxis]
        )
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, images=None, **kwargs
    ):
        batch_size = tf.shape(images)[0]
        input_height, input_width, _ = images.shape[1:]
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=images,
            dtype=self.compute_dtype,
        )
        boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]

        # values to translate the boxes by in the mosaic image
        mosaic_centers = transformations["mosaic_centers"]
        mosaic_centers_x = mosaic_centers[..., 0] * input_width
        mosaic_centers_y = mosaic_centers[..., 1] * input_height
        translate_x = tf.stack(
            [
                mosaic_centers_x - input_width,
                mosaic_centers_x,
                mosaic_centers_x - input_width,
                mosaic_centers_x,
            ],
            axis=-1,
        )
        translate_y = tf.stack(
            [
                mosaic_centers_y - input_height,
                mosaic_centers_y - input_height,
                mosaic_centers_y,
                mosaic_centers_y,
            ],
            axis=-1,
        )
        # updates bounding_boxes for one output mosaic
        permutation_order = transformations["permutation_order"]
        classes_for_mosaic = tf.gather(classes, permutation_order)
        boxes_for_mosaic = tf.gather(boxes, permutation_order)

        # stacking translate values such that the shape is (B, 4, 1, 4) or
        # (batch_size, num_images, broadcast dim, coordinates)
        translate_values = tf.stack(
            [translate_x, translate_y, translate_x, translate_y], axis=-1
        )
        translate_values = tf.expand_dims(translate_values, axis=2)
        # translating boxes
        boxes_for_mosaic = boxes_for_mosaic + translate_values
        boxes_for_mosaic = tf.reshape(boxes_for_mosaic, [batch_size, -1, 4])
        classes_for_mosaic = tf.reshape(classes_for_mosaic, [batch_size, -1])
        boxes_for_mosaic = {
            "boxes": boxes_for_mosaic,
            "classes": classes_for_mosaic,
        }
        boxes_for_mosaic = bounding_box.clip_to_image(
            boxes_for_mosaic,
            bounding_box_format="xyxy",
            images=images,
        )
        boxes_for_mosaic = bounding_box.convert_format(
            boxes_for_mosaic,
            source="xyxy",
            target=self.bounding_box_format,
            images=images,
            dtype=self.compute_dtype,
        )
        return boxes_for_mosaic

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        return super()._batch_augment(inputs)

    def call(self, inputs):
        _, metadata = self._format_inputs(inputs)
        if metadata[BATCHED] is not True:
            raise ValueError(
                "Mosaic received a single image to `call`. The "
                "layer relies on combining multiple examples, and as such "
                "will not behave as expected. Please call the layer with 4 "
                "or more samples."
            )
        return super().call(inputs=inputs)

    def _validate_inputs(self, inputs):
        images = inputs.get(IMAGES, None)
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        if images is None or (
            labels is None
            and bounding_boxes is None
            and segmentation_masks is None
        ):
            raise ValueError(
                "Mosaic expects inputs in a dictionary with format "
                '{"images": images, "labels": labels} or'
                '{"images": images, "bounding_boxes": bounding_boxes} or'
                '{"images": images, "segmentation_masks": masks}.'
                f"Got: inputs = {inputs}"
            )
        if labels is not None and not labels.dtype.is_floating:
            raise ValueError(
                f"Mosaic received labels with type {labels.dtype}. "
                "Labels must be of type float."
            )
        if bounding_boxes is not None and self.bounding_box_format is None:
            raise ValueError(
                "Mosaic received bounding boxes but no bounding_box_format. "
                "Please pass a bounding_box_format from the supported list."
            )

    def get_config(self):
        config = {
            "offset": self.offset,
            "bounding_box_format": self.bounding_box_format,
            "seed": self.seed,
        }
        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import tree

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import config
from keras_cv.src.backend import keras
from keras_cv.src.models.feature_extractor.clip.clip_processor_utils import (
    convert_inputs_to_list_of_tensor_segments,
)
from keras_cv.src.models.feature_extractor.clip.clip_processor_utils import (
    convert_to_backend_tensor_or_python_list,
)
from keras_cv.src.models.feature_extractor.clip.clip_tokenizer import (
    CLIPTokenizer,
)

try:
    import keras_nlp
    from keras_nlp.layers import StartEndPacker
except ImportError:
    keras_nlp = None
    StartEndPacker = None


@keras_cv_export("keras_cv.models.feature_extractor.CLIPProcessor")
class CLIPProcessor(keras.layers.Layer):
    """
    CLIPProcessor is a utility class that provides functionality for processing
    texts in the context of the CLIP (Contrastive Language-Image
    Pretraining) model.

    Args:
        input_resolution (int): The resolution of input images.
        vocabulary (str): string or dict, maps token to integer ids. If it is a
            string, it should be the file path to a json file.
        merges: string or list, contains the merge rule. If it is a string, it
            should be the file path to merge rules. The merge rule file should
            have one merge rule per line.

    """

    def __init__(self, vocabulary, merges, **kwargs):
        super().__init__(**kwargs)
        if keras_nlp is None:
            raise ValueError(
                "ClipTokenizer requires keras-nlp. Please install "
                "using pip `pip install -U keras-nlp && pip install -U keras`"
            )
        self.vocabulary = vocabulary
        self.merges = merges
        self.tokenizer = CLIPTokenizer(
            vocabulary=self.vocabulary,
            merges=self.merges,
        )
        self._convert_input_args = False
        self._allow_non_tensor_positional_args = True

    def build(self, input_shape):
        # Defer packer creation to `build()` so that we can be sure tokenizer
        # assets have loaded when restoring a saved model.
        self.packer = StartEndPacker(
            start_value=self.tokenizer.token_to_id("<|startoftext|>"),
            end_value=self.tokenizer.token_to_id("<|endoftext|>"),
            pad_value=None,
            sequence_length=77,
            return_padding_mask=True,
        )
        self.built = True

    def _process_texts(self, texts, context_length: int = 77):
        # Ensure the layer is built
        if not self.built:
            self.build(None)

        texts = convert_inputs_to_list_of_tensor_segments(texts)

        if len(texts) != 1:
            raise ValueError(
                "CLIP requires each input feature to contain only "
                f"one segment, but received {len(texts)}."
            )

        token_ids, padding_mask = self.packer(
            self.tokenizer(texts[0]),
            sequence_length=context_length,
            add_start_value=True,
            add_end_value=True,
        )
        return {"token_ids": token_ids, "padding_mask": padding_mask}

    def call(self, texts, context_length: int = 77):
        return self._process_texts(texts, context_length=context_length)

    def get_build_config(self):
        return None

    def __call__(self, *args, **kwargs):
        # Always place on CPU for preprocessing, to avoid expensive back and
        # forth copies to GPU before the trainable model.
        with tf.device("cpu"):
            outputs = super().__call__(*args, **kwargs)

            # Jax and Torch lack native string and ragged types.
            # If we are running on those backends and not running with tf.data
            # (we are outside a tf.function), we covert all ragged and string
            # tensor to pythonic types.
            is_tf_backend = config.backend() == "tensorflow"
            is_in_tf_graph = not tf.executing_eagerly()
            if not is_tf_backend and not is_in_tf_graph:
                outputs = tree.map_structure(
                    convert_to_backend_tensor_or_python_list, outputs
                )

        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary": self.vocabulary,
                "merges": self.merges,
            }
        )
        return config



# Copyright 2024 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""VideoClassifier Task presets."""

classifier_presets = {
    "videoswin_tiny_kinetics400_classifier": {
        "metadata": {
            "description": (
                "A tiny Video Swin architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "78.8% top1 and 93.6% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 28_158_070,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
    "videoswin_small_kinetics400_classifier": {
        "metadata": {
            "description": (
                "A small Video Swin architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "80.6% top1 and 94.5% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 49_816_678,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
    "videoswin_base_kinetics400_classifier": {
        "metadata": {
            "description": (
                "A base Video Swin architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "80.6% top1 and 94.6% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 89_065_688,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
    "videoswin_base_kinetics400_imagenet22k": {
        "metadata": {
            "description": (
                "A base Video Swin architecture. "
                "It is pretrained on ImageNet 22K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "82.7% top1 and 95.5% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 89_065_688,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
    "videoswin_base_kinetics600_imagenet22k": {
        "metadata": {
            "description": (
                "A base Video Swin architecture. "
                "It is pretrained on ImageNet 22K dataset, and "
                "trained on Kinetics 600 dataset. "
                "Published weight is capable of scoring "
                "84.0% top1 and 96.5% top5 accuracy on the "
                "Kinetics 600 dataset"
            ),
            "params": 89_270_688,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
    "videoswin_base_something_something_v2_classifier": {
        "metadata": {
            "description": (
                "A base Video Swin architecture. "
                "It is pretrained on Kinetics 400 dataset, and "
                "trained on Something Something V2 dataset. "
                "Published weight is capable of scoring "
                "69.6% top1 and 92.7% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 88_834_038,
            "official_name": "VideoClassifier",
            "path": "video_classifier",
        },
    },
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLIP presets."""

clip_presets = {
    "clip-vit-base-patch16": {
        "metadata": {
            "description": (
                "The model uses a ViT-B/16 Transformer architecture as an "
                "image encoder and uses a masked self-attention Transformer as "
                "a text encoder. These encoders are trained to maximize the "
                "similarity of (image, text) pairs via a contrastive loss. The "
                "model uses a patch size of 16 and input images of size (224, "
                "224)"
            ),
            "params": 149620737,
            "official_name": "CLIP",
            "path": "clip",
        },
        "kaggle_handle": "kaggle://keras/clip/keras/clip-vit-base-patch16/7",
    },
    "clip-vit-base-patch32": {
        "metadata": {
            "description": (
                "The model uses a ViT-B/32 Transformer architecture as an "
                "image encoder and uses a masked self-attention Transformer as "
                "a text encoder. These encoders are trained to maximize the "
                "similarity of (image, text) pairs via a contrastive loss.The "
                "model uses a patch size of 32 and input images of size (224, "
                "224)"
            ),
            "params": 151277313,
            "official_name": "CLIP",
            "path": "clip",
        },
        "kaggle_handle": "kaggle://keras/clip/keras/clip-vit-base-patch32/6",
    },
    "clip-vit-large-patch14": {
        "metadata": {
            "description": (
                "The model uses a ViT-L/14 Transformer architecture as an "
                "image encoder and uses a masked self-attention Transformer as "
                "a text encoder. These encoders are trained to maximize the "
                "similarity of (image, text) pairs via a contrastive loss.The "
                "model uses a patch size of 14 and input images of size (224, "
                "224)"
            ),
            "params": 427616513,
            "official_name": "CLIP",
            "path": "clip",
        },
        "kaggle_handle": "kaggle://keras/clip/keras/clip-vit-large-patch14/6",
    },
    "clip-vit-large-patch14-336": {
        "metadata": {
            "description": (
                "The model uses a ViT-L/14 Transformer architecture as an "
                "image encoder and uses a masked self-attention Transformer as "
                "a text encoder. These encoders are trained to maximize the "
                "similarity of (image, text) pairs via a contrastive loss.The "
                "model uses a patch size of 14 and input images of size (336, "
                "336)"
            ),
            "params": 427944193,
            "official_name": "CLIP",
            "path": "clip",
        },
        "kaggle_handle": "kaggle://keras/clip/keras/clip-vit-large-patch14-336/6",  # noqa: E501
    },
}



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.models.object_detection.retinanet.feature_pyramid import (
    FeaturePyramid,
)
from keras_cv.src.models.object_detection.retinanet.prediction_head import (
    PredictionHead,
)
from keras_cv.src.models.object_detection.retinanet.retinanet_label_encoder import (
    RetinaNetLabelEncoder as LabelEncoder,
)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.models.feature_extractor.QuickGELU")
class QuickGELU(keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def call(self, x):
        return x * ops.sigmoid(1.702 * x)


@keras_cv_export("keras_cv.models.feature_extractor.ResidualAttention")
class ResidualAttention(keras.layers.Layer):
    def __init__(
        self,
        proj_dim,
        num_heads,
        num_hidden_layers,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.proj_dim = proj_dim
        self.num_heads = num_heads
        self.num_hidden_layers = num_hidden_layers
        self.attn = CLIPAttention(
            self.proj_dim,
            self.num_heads,
            self.num_hidden_layers,
            name="multi_head_attention",
        )
        self.ln_1 = keras.layers.LayerNormalization(epsilon=1e-5, name="ln_1")
        self.mlp_dense_1 = keras.layers.Dense(
            self.proj_dim * 4,
            name="c_fc",
        )
        self.mlp_activation = QuickGELU(name="gelu")
        self.mlp_dense_2 = keras.layers.Dense(
            self.proj_dim,
            name="c_proj",
        )
        self.ln_2 = keras.layers.LayerNormalization(epsilon=1e-5, name="ln_2")

    def attention(self, x, causal_attention_mask=None, attention_mask=None):
        mask = None
        if causal_attention_mask is not None:
            mask = (
                ops.cast(causal_attention_mask, dtype=x.dtype)
                if causal_attention_mask is not None
                else None
            )
        if attention_mask is not None:
            attention_mask = (
                ops.cast(attention_mask, dtype=x.dtype)
                if attention_mask is not None
                else None
            )
            mask = ops.add(causal_attention_mask, attention_mask)

        return self.attn(
            x,
            attention_mask=mask,
        )[0]

    def build(self, input_shape):
        super().build(input_shape)
        self.attn.build(None)
        self.ln_1.build([None, None, self.proj_dim])
        self.mlp_dense_1.build([None, None, self.proj_dim])
        self.mlp_dense_2.build([None, None, self.proj_dim * 4])
        self.ln_2.build([None, None, self.proj_dim])

    def call(self, x, causal_attention_mask=None, attention_mask=None):
        residual = x
        x = self.ln_1(x)
        x = self.attention(
            x,
            causal_attention_mask=causal_attention_mask,
            attention_mask=attention_mask,
        )
        x = x + residual
        residual = x
        x = self.mlp_dense_1(self.ln_2(residual))
        x = self.mlp_activation(x)
        x = self.mlp_dense_2(x)
        x = residual + x
        return x

    def compute_output_shape(self, inputs_shape):
        return inputs_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "proj_dim": self.proj_dim,
                "num_heads": self.num_heads,
                "num_hidden_layers": self.num_hidden_layers,
            }
        )
        return config


@keras_cv_export("keras_cv.models.feature_extractor.CLIPEncoder")
class CLIPEncoder(keras.layers.Layer):
    def __init__(self, width, num_layers, heads, **kwargs):
        super().__init__(**kwargs)
        self.width = width
        self.num_layers = num_layers
        self.heads = heads
        self.resblocks = [
            ResidualAttention(
                self.width,
                self.heads,
                self.num_layers,
            )
            for _ in range(self.num_layers)
        ]

    def build(self, input_shape):
        for block in self.resblocks:
            block.build(input_shape)
        self.built = True

    def compute_output_shape(self, input_shape):
        for block in self.resblocks:
            input_shape = block.compute_output_shape(input_shape)
        return input_shape

    def call(
        self,
        x,
        causal_attention_mask=None,
        attention_mask=None,
    ):
        for block in self.resblocks:
            x = block(
                x,
                causal_attention_mask=causal_attention_mask,
                attention_mask=attention_mask,
            )
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width": self.width,
                "num_layers": self.num_layers,
                "heads": self.heads,
            }
        )
        return config


@keras_cv_export("keras_cv.models.feature_extractor.CLIPAttention")
class CLIPAttention(keras.layers.Layer):
    """
    Adapted from https://github.com/huggingface/transformers/blob/main/src/transformers/models/clip/modeling_clip.py # noqa: E501
    """

    def __init__(
        self, proj_dim, num_heads, num_hidden_layers, dropout=0.0, **kwargs
    ):
        super().__init__(**kwargs)

        self.proj_dim = proj_dim
        self.num_heads = num_heads
        self.num_hidden_layers = num_hidden_layers
        self.dropout = dropout
        self.head_dim = self.proj_dim // self.num_heads
        if self.head_dim * self.num_heads != self.proj_dim:
            raise ValueError(
                f"proj_dim must be divisible by num_heads (got `proj_dim`"
                f": {self.proj_dim} and `num_heads`:"
                f" {self.num_heads})."
            )

        self.scale = self.head_dim**-0.5
        self.q_proj = keras.layers.Dense(
            units=self.proj_dim,
            name="q_proj",
        )
        self.k_proj = keras.layers.Dense(
            units=self.proj_dim,
            name="k_proj",
        )
        self.v_proj = keras.layers.Dense(
            units=self.proj_dim,
            name="v_proj",
        )
        self.out_proj = keras.layers.Dense(
            units=self.proj_dim,
            name="out_proj",
        )

    def build(self, input_shape):
        super().build(input_shape)
        self.q_proj.build([None, None, self.proj_dim])
        self.k_proj.build([None, None, self.proj_dim])
        self.v_proj.build([None, None, self.proj_dim])
        self.out_proj.build([None, None, self.proj_dim])

    def _transpose_for_scores(self, tensor, batch_size):
        """
        Adapted from https://github.com/huggingface/transformers/blob/8e164c5400b7b413c7b8fb32e35132001effc970/src/transformers/models/bert/modeling_tf_bert.py#L252 # noqa: E501
        """
        # [batch_size, seq_len, all_head_dim] ->
        # [batch_size, seq_len, num_heads, head_dim]
        tensor = ops.reshape(
            tensor, (batch_size, -1, self.num_heads, self.head_dim)
        )
        # [batch_size, seq_len, num_heads, head_dim] ->
        # [batch_size, num_heads, seq_len, head_dim]
        return ops.transpose(tensor, axes=[0, 2, 1, 3])

    def call(
        self,
        x,
        attention_mask=None,
        output_attentions=None,
        training=False,
    ):
        batch_size = ops.shape(x)[0]
        mixed_query_layer = self.q_proj(inputs=x)
        mixed_key_layer = self.k_proj(inputs=x)
        mixed_value_layer = self.v_proj(inputs=x)
        query_layer = self._transpose_for_scores(mixed_query_layer, batch_size)
        key_layer = self._transpose_for_scores(mixed_key_layer, batch_size)
        value_layer = self._transpose_for_scores(mixed_value_layer, batch_size)

        # Scaled dot product between key and query = raw attention scores.
        attention_scores = ops.matmul(
            query_layer, ops.transpose(key_layer, axes=[0, 1, 3, 2])
        )
        dk = ops.cast(ops.sqrt(self.head_dim), dtype=attention_scores.dtype)
        attention_scores = ops.divide(
            attention_scores, dk
        )  # (batch_size, num_heads, seq_len_q, seq_len_k)

        if attention_mask is not None:
            # Apply the attention mask (precomputed for all layers in the
            # call() function)
            attention_scores = ops.add(attention_scores, attention_mask)

        # Normalize the attention scores to probabilities.
        attention_probs = ops.softmax(attention_scores, axis=-1)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        dropout_attention_probs = keras.layers.Dropout(self.dropout)(
            inputs=attention_probs, training=training
        )

        attn_output = ops.matmul(dropout_attention_probs, value_layer)
        attn_output = ops.transpose(attn_output, axes=[0, 2, 1, 3])

        # (batch_size, seq_len_q, proj_dim)
        attn_output = ops.reshape(attn_output, (batch_size, -1, self.proj_dim))

        attn_output = self.out_proj(attn_output, training=training)
        outputs = (
            (attn_output, attention_probs)
            if output_attentions
            else (attn_output,)
        )

        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "proj_dim": self.proj_dim,
                "num_heads": self.num_heads,
                "num_hidden_layers": self.num_hidden_layers,
                "dropout": self.dropout,
            }
        )
        return config



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""EfficientNet Lite backbone model.

Reference:
    - [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks](https://arxiv.org/abs/1905.11946)
        (ICML 2019)
    - [Based on the original EfficientNet Lite's](https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet/lite)
"""  # noqa: E501

import copy
import math

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.models import utils
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.utils.python_utils import classproperty

BN_AXIS = 3


@keras_cv_export("keras_cv.models.EfficientNetLiteBackbone")
@keras.saving.register_keras_serializable(package="keras_cv.models")
class EfficientNetLiteBackbone(Backbone):
    """Instantiates the EfficientNetLite architecture using given scaling
    coefficients.

    Reference:
    - [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks](https://arxiv.org/abs/1905.11946)
        (ICML 2019)
    - [Based on the original EfficientNet Lite's](https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet/lite)

    Args:
        include_rescaling: whether to rescale the inputs. If set to True,
            inputs will be passed through a `Rescaling(1/255.0)` layer.
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections. The
            default value is set to 0.2.
        depth_divisor: integer, a unit of network width. The default value
            is set to 8.
        activation: activation function.
        input_shape: optional shape tuple,
            It should have exactly 3 inputs channels.
        input_tensor: optional Keras tensor (i.e. output of `keras.layers.Input()`)
            to use as image input for the model.

    Example:
    ```python
    # Construct an EfficientNetLite from a preset:
    efficientnet = models.EfficientNetLiteBackbone.from_preset(
        "efficientnetlite_b0"
    )
    images = np.ones((1, 256, 256, 3))
    outputs = efficientnet.predict(images)

    # Alternatively, you can also customize the EfficientNetLite architecture:
    model = EfficientNetLiteBackbone(
        stackwise_kernel_sizes=[3, 3, 5, 3, 5, 5, 3],
        stackwise_num_repeats=[1, 2, 2, 3, 3, 4, 1],
        stackwise_input_filters=[32, 16, 24, 40, 80, 112, 192],
        stackwise_output_filters=[16, 24, 40, 80, 112, 192, 320],
        stackwise_expansion_ratios=[1, 6, 6, 6, 6, 6, 6],
        stackwise_strides=[1, 2, 2, 2, 1, 2, 1],
        width_coefficient=1.0,
        depth_coefficient=1.0,
        include_rescaling=False,
    )
    images = np.ones((1, 256, 256, 3))
    outputs = model.predict(images)
    ```
    """  # noqa: E501

    def __init__(
        self,
        *,
        include_rescaling,
        width_coefficient,
        depth_coefficient,
        stackwise_kernel_sizes,
        stackwise_num_repeats,
        stackwise_input_filters,
        stackwise_output_filters,
        stackwise_expansion_ratios,
        stackwise_strides,
        dropout_rate=0.2,
        drop_connect_rate=0.2,
        depth_divisor=8,
        input_shape=(None, None, 3),
        input_tensor=None,
        activation="relu6",
        **kwargs,
    ):
        img_input = utils.parse_model_inputs(input_shape, input_tensor)

        # Build stem
        x = img_input

        if include_rescaling:
            # Use common rescaling strategy across keras_cv
            x = keras.layers.Rescaling(1.0 / 255.0)(x)

        x = keras.layers.ZeroPadding2D(
            padding=utils.correct_pad_downsample(x, 3), name="stem_conv_pad"
        )(x)
        x = keras.layers.Conv2D(
            32,
            3,
            strides=2,
            padding="valid",
            use_bias=False,
            kernel_initializer=conv_kernel_initializer(),
            name="stem_conv",
        )(x)
        x = keras.layers.BatchNormalization(axis=BN_AXIS, name="stem_bn")(x)
        x = keras.layers.Activation(activation, name="stem_activation")(x)

        # Build blocks
        block_id = 0
        blocks = float(sum(stackwise_num_repeats))

        pyramid_level_inputs = []

        for i in range(len(stackwise_kernel_sizes)):
            num_repeats = stackwise_num_repeats[i]
            input_filters = stackwise_input_filters[i]
            output_filters = stackwise_output_filters[i]
            # Update block input and output filters based on depth multiplier.
            input_filters = round_filters(
                filters=input_filters,
                width_coefficient=width_coefficient,
                depth_divisor=depth_divisor,
            )
            output_filters = round_filters(
                filters=output_filters,
                width_coefficient=width_coefficient,
                depth_divisor=depth_divisor,
            )

            if i == 0 or i == (len(stackwise_kernel_sizes) - 1):
                repeats = num_repeats
            else:
                repeats = round_repeats(
                    repeats=num_repeats,
                    depth_coefficient=depth_coefficient,
                )
            strides = stackwise_strides[i]

            for j in range(repeats):
                # The first block needs to take care of stride and filter size
                # increase.
                if j > 0:
                    strides = 1
                    input_filters = output_filters

                if strides != 1:
                    pyramid_level_inputs.append(utils.get_tensor_input_name(x))

                # 97 is the start of the lowercase alphabet.
                letter_identifier = chr(j + 97)
                x = apply_efficient_net_lite_block(
                    inputs=x,
                    filters_in=input_filters,
                    filters_out=output_filters,
                    kernel_size=stackwise_kernel_sizes[i],
                    strides=strides,
                    expand_ratio=stackwise_expansion_ratios[i],
                    activation=activation,
                    dropout_rate=drop_connect_rate * block_id / blocks,
                    name="block{}{}_".format(i + 1, letter_identifier),
                )
                block_id += 1

        # Build top
        x = keras.layers.Conv2D(
            1280,
            1,
            padding="same",
            use_bias=False,
            kernel_initializer=conv_kernel_initializer(),
            name="top_conv",
        )(x)
        x = keras.layers.BatchNormalization(axis=BN_AXIS, name="top_bn")(x)
        x = keras.layers.Activation(activation, name="top_activation")(x)

        pyramid_level_inputs.append(utils.get_tensor_input_name(x))

        # Create model.
        super().__init__(inputs=img_input, outputs=x, **kwargs)

        self.include_rescaling = include_rescaling
        self.width_coefficient = width_coefficient
        self.depth_coefficient = depth_coefficient
        self.dropout_rate = dropout_rate
        self.drop_connect_rate = drop_connect_rate
        self.depth_divisor = depth_divisor
        self.activation = activation
        self.input_tensor = input_tensor
        self.pyramid_level_inputs = {
            f"P{i + 1}": name for i, name in enumerate(pyramid_level_inputs)
        }
        self.stackwise_kernel_sizes = stackwise_kernel_sizes
        self.stackwise_num_repeats = stackwise_num_repeats
        self.stackwise_input_filters = stackwise_input_filters
        self.stackwise_output_filters = stackwise_output_filters
        self.stackwise_expansion_ratios = stackwise_expansion_ratios
        self.stackwise_strides = stackwise_strides

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "include_rescaling": self.include_rescaling,
                "width_coefficient": self.width_coefficient,
                "depth_coefficient": self.depth_coefficient,
                "dropout_rate": self.dropout_rate,
                "drop_connect_rate": self.drop_connect_rate,
                "depth_divisor": self.depth_divisor,
                "activation": self.activation,
                "input_tensor": self.input_tensor,
                "input_shape": self.input_shape[1:],
                "stackwise_kernel_sizes": self.stackwise_kernel_sizes,
                "stackwise_num_repeats": self.stackwise_num_repeats,
                "stackwise_input_filters": self.stackwise_input_filters,
                "stackwise_output_filters": self.stackwise_output_filters,
                "stackwise_expansion_ratios": self.stackwise_expansion_ratios,
                "stackwise_strides": self.stackwise_strides,
            }
        )
        return config

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)


def conv_kernel_initializer(scale=2.0):
    return keras.initializers.VarianceScaling(
        scale=scale, mode="fan_out", distribution="truncated_normal"
    )


def round_filters(filters, depth_divisor, width_coefficient):
    """Round number of filters based on depth multiplier."""
    filters *= width_coefficient
    new_filters = max(
        depth_divisor,
        int(filters + depth_divisor / 2) // depth_divisor * depth_divisor,
    )
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats, depth_coefficient):
    """Round number of repeats based on depth multiplier."""
    return int(math.ceil(depth_coefficient * repeats))


def apply_efficient_net_lite_block(
    inputs,
    activation="relu6",
    dropout_rate=0.0,
    name=None,
    filters_in=32,
    filters_out=16,
    kernel_size=3,
    strides=1,
    expand_ratio=1,
):
    """An inverted residual block, without SE phase.

    Args:
        inputs: input tensor.
        activation: activation function.
        dropout_rate: float between 0 and 1, fraction of the input units to drop.
        name: string, block label.
        filters_in: integer, the number of input filters.
        filters_out: integer, the number of output filters.
        kernel_size: integer, the dimension of the convolution window.
        strides: integer, the stride of the convolution.
        expand_ratio: integer, scaling coefficient for the input filters.

    Returns:
        output tensor for the block.
    """  # noqa: E501
    if name is None:
        name = f"block_{keras.backend.get_uid('block_')}_"

    # Expansion phase
    filters = filters_in * expand_ratio
    if expand_ratio != 1:
        x = keras.layers.Conv2D(
            filters,
            1,
            padding="same",
            use_bias=False,
            kernel_initializer=conv_kernel_initializer(),
            name=name + "expand_conv",
        )(inputs)
        x = keras.layers.BatchNormalization(
            axis=BN_AXIS, name=name + "expand_bn"
        )(x)
        x = keras.layers.Activation(
            activation, name=name + "expand_activation"
        )(x)
    else:
        x = inputs

    # Depthwise Convolution
    if strides == 2:
        x = keras.layers.ZeroPadding2D(
            padding=utils.correct_pad_downsample(x, kernel_size),
            name=name + "dwconv_pad",
        )(x)
        conv_pad = "valid"
    else:
        conv_pad = "same"
    x = keras.layers.DepthwiseConv2D(
        kernel_size,
        strides=strides,
        padding=conv_pad,
        use_bias=False,
        depthwise_initializer=conv_kernel_initializer(),
        name=name + "dwconv",
    )(x)
    x = keras.layers.BatchNormalization(axis=BN_AXIS, name=name + "bn")(x)
    x = keras.layers.Activation(activation, name=name + "activation")(x)

    # Output phase
    x = keras.layers.Conv2D(
        filters_out,
        1,
        padding="same",
        use_bias=False,
        kernel_initializer=conv_kernel_initializer(),
        name=name + "project_conv",
    )(x)
    x = keras.layers.BatchNormalization(axis=BN_AXIS, name=name + "project_bn")(
        x
    )
    if strides == 1 and filters_in == filters_out:
        if dropout_rate > 0:
            x = keras.layers.Dropout(
                dropout_rate, noise_shape=(None, 1, 1, 1), name=name + "drop"
            )(x)
        x = keras.layers.Add(name=name + "add")([x, inputs])
    return x



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src import core
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers import preprocessing as cv_preprocessing
from keras_cv.src.layers.preprocessing.random_augmentation_pipeline import (
    RandomAugmentationPipeline,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils


@keras_cv_export("keras_cv.layers.RandAugment")
class RandAugment(RandomAugmentationPipeline):
    """RandAugment performs the Rand Augment operation on input images.

    This layer can be thought of as an all-in-one image augmentation layer. The
    policy implemented by this layer has been benchmarked extensively and is
    effective on a wide variety of datasets.

    The policy operates as follows:

    For each augmentation in the range `[0, augmentations_per_image]`,
    the policy selects a random operation from a list of operations.
    It then samples a random number and if that number is less than
    `rate` applies it to the given image.

    References:
        - [RandAugment](https://arxiv.org/abs/1909.13719)

    Args:
        value_range: the range of values the incoming images will have.
            Represented as a two number tuple written [low, high].
            This is typically either `[0, 1]` or `[0, 255]` depending
            on how your preprocessing pipeline is set up.
        augmentations_per_image: the number of layers to use in the rand augment
            policy, defaults to `3`.
        magnitude: magnitude is the mean of the normal distribution used to
            sample the magnitude used for each data augmentation. Magnitude
            should be a float in the range `[0, 1]`. A magnitude of `0`
            indicates that the augmentations are as weak as possible (not
            recommended), while a value of `1.0` implies use of the strongest
            possible augmentation. All magnitudes are clipped to the range
            `[0, 1]` after sampling. Defaults to `0.5`.
        magnitude_stddev: the standard deviation to use when drawing values for
            the perturbations. Keep in mind magnitude will still be clipped to
            the range `[0, 1]` after samples are drawn from the normal
            distribution. Defaults to `0.15`.
        rate: the rate at which to apply each augmentation. This parameter is
            applied on a per-distortion layer, per image. Should be in the range
            `[0, 1]`. To reproduce the original RandAugment paper results, set
            this to `10/11`. The original `RandAugment` paper includes an
            Identity transform. By setting the rate to 10/11 in our
            implementation, the behavior is identical to sampling an Identity
            augmentation 10/11th of the time. Defaults to `1.0`.
        geometric: whether to include geometric augmentations. This
            should be set to False when performing object detection. Defaults to
            True.
    Example:
    ```python
    (x_test, y_test), _ = keras.datasets.cifar10.load_data()
    rand_augment = keras_cv.layers.RandAugment(
        value_range=(0, 255), augmentations_per_image=3, magnitude=0.5
    )
    x_test = rand_augment(x_test)
    ```
    """

    def __init__(
        self,
        value_range,
        augmentations_per_image=3,
        magnitude=0.5,
        magnitude_stddev=0.15,
        rate=10 / 11,
        geometric=True,
        seed=None,
        **kwargs,
    ):
        # As an optimization RandAugment makes all internal layers use (0, 255)
        # and we handle range transformation at the _augment level.
        if magnitude < 0.0 or magnitude > 1:
            raise ValueError(
                "`magnitude` must be in the range [0, 1], got "
                f"`magnitude={magnitude}`"
            )
        if magnitude_stddev < 0.0 or magnitude_stddev > 1:
            raise ValueError(
                "`magnitude_stddev` must be in the range [0, 1], got "
                f"`magnitude_stddev={magnitude}`"
            )

        super().__init__(
            layers=RandAugment.get_standard_policy(
                (0, 255),
                magnitude,
                magnitude_stddev,
                geometric=geometric,
                seed=seed,
            ),
            augmentations_per_image=augmentations_per_image,
            rate=rate,
            **kwargs,
            seed=seed,
        )
        self.magnitude = float(magnitude)
        self.value_range = value_range
        self.seed = seed
        self.geometric = geometric
        self.magnitude_stddev = float(magnitude_stddev)

    def _augment(self, sample):
        sample["images"] = preprocessing_utils.transform_value_range(
            sample["images"], self.value_range, (0, 255)
        )
        result = super()._augment(sample)
        result["images"] = preprocessing_utils.transform_value_range(
            result["images"], (0, 255), self.value_range
        )
        result["images"]
        return result

    @staticmethod
    def get_standard_policy(
        value_range, magnitude, magnitude_stddev, geometric=True, seed=None
    ):
        policy = create_rand_augment_policy(magnitude, magnitude_stddev)

        auto_contrast = cv_preprocessing.AutoContrast(
            **policy["auto_contrast"], value_range=value_range, seed=seed
        )
        equalize = cv_preprocessing.Equalization(
            **policy["equalize"], value_range=value_range, seed=seed
        )

        solarize = cv_preprocessing.Solarization(
            **policy["solarize"], value_range=value_range, seed=seed
        )

        color = cv_preprocessing.RandomColorDegeneration(
            **policy["color"], seed=seed
        )
        contrast = cv_preprocessing.RandomContrast(
            **policy["contrast"], value_range=value_range, seed=seed
        )
        brightness = cv_preprocessing.RandomBrightness(
            **policy["brightness"], value_range=value_range, seed=seed
        )

        layers = [
            auto_contrast,
            equalize,
            solarize,
            color,
            contrast,
            brightness,
        ]

        if geometric:
            shear_x = cv_preprocessing.RandomShear(
                **policy["shear_x"], seed=seed
            )
            shear_y = cv_preprocessing.RandomShear(
                **policy["shear_y"], seed=seed
            )
            translate_x = cv_preprocessing.RandomTranslation(
                **policy["translate_x"], seed=seed
            )
            translate_y = cv_preprocessing.RandomTranslation(
                **policy["translate_y"], seed=seed
            )
            layers += [shear_x, shear_y, translate_x, translate_y]
        return layers

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "augmentations_per_image": self.augmentations_per_image,
                "magnitude": self.magnitude,
                "magnitude_stddev": self.magnitude_stddev,
                "rate": self.rate,
                "geometric": self.geometric,
                "seed": self.seed,
            }
        )
        # layers is recreated in the constructor
        del config["layers"]
        return config


def auto_contrast_policy(magnitude, magnitude_stddev):
    return {}


def equalize_policy(magnitude, magnitude_stddev):
    return {}


def solarize_policy(magnitude, magnitude_stddev):
    # We cap additions at 110, because if we add more than 110 we will be nearly
    # nullifying the information contained in the image, making the model train
    # on noise
    maximum_addition_value = 110
    addition_factor = core.NormalFactorSampler(
        mean=magnitude * maximum_addition_value,
        stddev=magnitude_stddev * maximum_addition_value,
        min_value=0,
        max_value=maximum_addition_value,
    )
    threshold_factor = core.NormalFactorSampler(
        mean=(255 - (magnitude * 255)),
        stddev=(magnitude_stddev * 255),
        min_value=0,
        max_value=255,
    )
    return {
        "addition_factor": addition_factor,
        "threshold_factor": threshold_factor,
    }


def color_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"factor": factor}


def contrast_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomContrast with `factor`?
    # RandomContrast layer errors when factor=0
    factor = max(magnitude, 0.001)
    return {"factor": factor}


def brightness_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomBrightness with `factor`?
    return {"factor": magnitude}


def shear_x_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"x_factor": factor, "y_factor": 0}


def shear_y_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"x_factor": 0, "y_factor": factor}


def translate_x_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomTranslation with `factor`?
    return {"width_factor": magnitude, "height_factor": 0}


def translate_y_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomTranslation with `factor`?
    return {"width_factor": 0, "height_factor": magnitude}


POLICY_PAIRS = {
    "auto_contrast": auto_contrast_policy,
    "equalize": equalize_policy,
    "solarize": solarize_policy,
    "color": color_policy,
    "contrast": contrast_policy,
    "brightness": brightness_policy,
    "shear_x": shear_x_policy,
    "shear_y": shear_y_policy,
    "translate_x": translate_x_policy,
    "translate_y": translate_y_policy,
}


def create_rand_augment_policy(magnitude, magnitude_stddev):
    result = {}
    for name, policy_fn in POLICY_PAIRS.items():
        result[name] = policy_fn(magnitude, magnitude_stddev)
    return result



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tensorflow as tf
from tensorflow import keras

import keras_cv.src.layers as cv_layers
from keras_cv.src import bounding_box


class YoloXPredictionDecoder(keras.layers.Layer):
    """Decodes the predictions from YoloX head.

    This layer is similar to the decoding code in `YoloX.compute_losses`. This
    is followed by a bounding box suppression layer.

    Arguments:
        bounding_box_format:  The format of bounding boxes of input dataset.
            Refer to https://keras.io/api/keras_cv/bounding_box/formats/
            for more details on supported bounding box formats.
        num_classes: The number of classes to be considered for the
            classification head.
        suppression_layer: A `keras.layers.Layer` that follows the same API
            signature of the `keras_cv.layers.MultiClassNonMaxSuppression`
            layer. This layer should perform a suppression operation such as Non
            Max Suppression, or Soft Non-Max Suppression.
    """

    def __init__(
        self, bounding_box_format, num_classes, suppression_layer=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.bounding_box_format = bounding_box_format
        self.num_classes = num_classes

        self.suppression_layer = (
            suppression_layer
            or cv_layers.MultiClassNonMaxSuppression(
                bounding_box_format=bounding_box_format,
                from_logits=False,
                confidence_threshold=0.01,
                iou_threshold=0.65,
                max_detections=100,
                max_detections_per_class=100,
            )
        )
        if (
            self.suppression_layer.bounding_box_format
            != self.bounding_box_format
        ):
            raise ValueError(
                "`suppression_layer` must have the same `bounding_box_format` "
                "as the `YoloXPredictionDecoder()` layer. "
                "Received `YoloXPredictionDecoder.bounding_box_format="
                f"{self.bounding_box_format}`, "
                f"`suppression_layer={suppression_layer}`."
            )
        self.built = True

    def call(self, images, predictions):
        image_shape = tf.cast(tf.shape(images), dtype=self.compute_dtype)[1:-1]

        batch_size = tf.shape(predictions[0])[0]

        grids = []
        strides = []

        shapes = [x.shape[1:3] for x in predictions]

        # 5 + self.num_classes is a concatenation of bounding boxes (length=4)
        # + objectness score (length=1) + num_classes
        # this reshape is simply collapsing axes 1 and 2 of x into a single
        # dimension
        predictions = [
            tf.reshape(x, [batch_size, -1, 5 + self.num_classes])
            for x in predictions
        ]
        predictions = tf.cast(
            tf.concat(predictions, axis=1), dtype=self.compute_dtype
        )
        predictions_shape = tf.cast(
            tf.shape(predictions), dtype=self.compute_dtype
        )

        for i in range(len(shapes)):
            shape_x, shape_y = shapes[i]
            grid_x, grid_y = tf.meshgrid(tf.range(shape_y), tf.range(shape_x))
            grid = tf.reshape(tf.stack((grid_x, grid_y), 2), (1, -1, 2))
            shape = grid.shape[:2]

            grids.append(tf.cast(grid, self.compute_dtype))
            strides.append(
                tf.ones((shape[0], shape[1], 1))
                * image_shape[0]
                / tf.cast(shape_x, self.compute_dtype)
            )

        grids = tf.concat(grids, axis=1)
        strides = tf.concat(strides, axis=1)

        box_xy = tf.expand_dims(
            (predictions[..., :2] + grids) * strides / image_shape, axis=-2
        )
        box_xy = tf.broadcast_to(
            box_xy, [batch_size, predictions_shape[1], self.num_classes, 2]
        )
        box_wh = tf.expand_dims(
            tf.exp(predictions[..., 2:4]) * strides / image_shape, axis=-2
        )
        box_wh = tf.broadcast_to(
            box_wh, [batch_size, predictions_shape[1], self.num_classes, 2]
        )

        box_confidence = tf.math.sigmoid(predictions[..., 4:5])
        box_class_probs = tf.math.sigmoid(predictions[..., 5:])

        # create and broadcast classes for every box before nms
        box_classes = tf.expand_dims(
            tf.range(self.num_classes, dtype=self.compute_dtype), axis=-1
        )
        box_classes = tf.broadcast_to(
            box_classes, [batch_size, predictions_shape[1], self.num_classes, 1]
        )

        box_scores = tf.expand_dims(box_confidence * box_class_probs, axis=-1)

        outputs = tf.concat([box_xy, box_wh, box_classes, box_scores], axis=-1)
        outputs = tf.reshape(outputs, [batch_size, -1, 6])

        outputs = {
            "boxes": outputs[..., :4],
            "classes": outputs[..., 4],
            "confidence": outputs[..., 5],
        }

        # this conversion is rel_center_xywh to rel_xywh
        # small workaround because rel_center_xywh isn't supported yet
        outputs = bounding_box.convert_format(
            outputs,
            source="center_xywh",
            target="xywh",
            images=images,
        )
        outputs = bounding_box.convert_format(
            outputs,
            source="rel_xywh",
            target=self.suppression_layer.bounding_box_format,
            images=images,
        )

        # preparing the predictions for TF NMS op
        class_predictions = tf.cast(outputs["classes"], tf.int32)
        class_predictions = tf.one_hot(class_predictions, self.num_classes)

        scores = (
            tf.expand_dims(outputs["confidence"], axis=-1) * class_predictions
        )

        return self.suppression_layer(outputs["boxes"], scores)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import cv2
except:
    cv2 = None

import numpy as np

from keras_cv.src import bounding_box
from keras_cv.src import utils
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.utils import assert_cv2_installed


@keras_cv_export("keras_cv.visualization.draw_bounding_boxes")
def draw_bounding_boxes(
    images,
    bounding_boxes,
    color,
    bounding_box_format,
    line_thickness=1,
    text_thickness=1,
    font_scale=1.0,
    class_mapping=None,
):
    """Internal utility to draw bounding boxes on the target image.

    Accepts a batch of images and batch of bounding boxes. The function draws
    the bounding boxes onto the image, and returns a new image tensor with the
    annotated images. This API is intentionally not exported, and is considered
    an implementation detail.

    Args:
        images: a batch Tensor of images to plot bounding boxes onto.
        bounding_boxes: a Tensor of batched bounding boxes to plot onto the
            provided images.
        color: the color in which to plot the bounding boxes
        bounding_box_format: The format of bounding boxes to plot onto the
            images. Refer
            [to the keras.io docs](https://keras.io/api/keras_cv/bounding_box/formats/)
            for more details on supported bounding box formats.
        line_thickness: (Optional) line_thickness for the box and text labels.
            Defaults to 2.
        text_thickness: (Optional) the thickness for the text, defaults to
            `1.0`.
        font_scale: (Optional) scale of font to draw in, defaults to `1.0`.
        class_mapping: (Optional) dictionary from class ID to class label.

    Returns:
        the input `images` with provided bounding boxes plotted on top of them
    """  # noqa: E501
    assert_cv2_installed("draw_bounding_boxes")
    bounding_boxes = bounding_box.convert_format(
        bounding_boxes, source=bounding_box_format, target="xyxy", images=images
    )
    text_thickness = text_thickness or line_thickness

    bounding_boxes["boxes"] = utils.to_numpy(bounding_boxes["boxes"])
    bounding_boxes["classes"] = utils.to_numpy(bounding_boxes["classes"])
    images = utils.to_numpy(images)
    image_width = images.shape[-2]
    outline_factor = image_width // 100

    class_mapping = class_mapping or {}
    result = []

    if len(images.shape) != 4:
        raise ValueError(
            "Images must be a batched np-like with elements of shape "
            "(height, width, 3)"
        )

    for i in range(images.shape[0]):
        bounding_box_batch = {
            "boxes": bounding_boxes["boxes"][i],
            "classes": bounding_boxes["classes"][i],
        }
        if "confidence" in bounding_boxes:
            bounding_box_batch["confidence"] = bounding_boxes["confidence"][i]

        image = utils.to_numpy(images[i]).astype("uint8")
        for b_id in range(bounding_box_batch["boxes"].shape[0]):
            x, y, x2, y2 = bounding_box_batch["boxes"][b_id].astype(int)
            class_id = bounding_box_batch["classes"][b_id].astype(int)
            confidence = bounding_box_batch.get("confidence", None)

            if class_id == -1:
                continue
            # force conversion back to contiguous array
            x, y, x2, y2 = int(x), int(y), int(x2), int(y2)
            cv2.rectangle(
                image,
                (x, y),
                (x2, y2),
                (0, 0, 0, 0.5),
                line_thickness + outline_factor,
            )
            cv2.rectangle(image, (x, y), (x2, y2), color, line_thickness)
            class_id = int(class_id)

            if class_id in class_mapping:
                label = class_mapping[class_id]
                if confidence is not None:
                    label = f"{label} | {confidence[b_id]:.2f}"

                x, y = _find_text_location(
                    x, y, font_scale, line_thickness, outline_factor
                )
                cv2.putText(
                    image,
                    label,
                    (x, y),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    (0, 0, 0, 0.5),
                    text_thickness + outline_factor,
                )
                cv2.putText(
                    image,
                    label,
                    (x, y),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    font_scale,
                    color,
                    text_thickness,
                )
        result.append(image)
    return np.array(result).astype(int)


def _find_text_location(x, y, font_scale, line_thickness, outline_factor):
    font_height = int(font_scale * 12)
    target_y = y - int(8 + outline_factor)
    if target_y - (2 * font_height) > 0:
        return x, y - int(8 + outline_factor)

    line_offset = line_thickness + outline_factor
    static_offset = 3

    return (
        x + outline_factor + static_offset,
        y + (2 * font_height) + line_offset + static_offset,
    )



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for keypoint transformation."""
import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export

H_AXIS = -3
W_AXIS = -2


@keras_cv_export("keras_cv.keypoint.filter_out_of_image")
def filter_out_of_image(keypoints, image):
    """Discards keypoints if falling outside of the image.

    Args:
      keypoints: a, possibly ragged, 2D (ungrouped), 3D (grouped)
        keypoint data in the 'xy' format.
      image: a 3D tensor in the HWC format.

    Returns:
      tf.RaggedTensor: a 2D or 3D ragged tensor with at least one
        ragged rank containing only keypoint in the image.
    """

    image_shape = tf.cast(tf.shape(image), keypoints.dtype)
    mask = tf.math.logical_and(
        tf.math.logical_and(
            keypoints[..., 0] >= 0, keypoints[..., 0] < image_shape[W_AXIS]
        ),
        tf.math.logical_and(
            keypoints[..., 1] >= 0, keypoints[..., 1] < image_shape[H_AXIS]
        ),
    )
    masked = tf.ragged.boolean_mask(keypoints, mask)
    if isinstance(masked, tf.RaggedTensor):
        return masked
    return tf.RaggedTensor.from_tensor(masked)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers.object_detection_3d import voxel_utils

EPSILON = 1e-4
VOXEL_FEATURE_MIN = -1000


def compute_point_voxel_id(point_voxel_xyz, voxel_spatial_size):
    """Computes point voxel IDs.

    Args:
      point_voxel_xyz: [B, N, dim] voxel coordinates for each point
      voxel_spatial_size: voxel spatial size

    Returns:
      point_voxel_id: [B, N] unique ID of each voxel.
    """
    batch_size, _, dim = list(point_voxel_xyz.shape)
    if batch_size is None:
        batch_size = ops.shape(point_voxel_xyz)[0]
    assert dim == len(voxel_spatial_size), f"{point_voxel_xyz.shape}"

    voxel_spatial_size_prod = [
        np.prod(voxel_spatial_size[i:]).item() for i in range(dim)
    ]
    voxel_spatial_size_prod_shift = voxel_spatial_size_prod[1:] + [1]
    point_voxel_xyz_multiplied = point_voxel_xyz * ops.array(
        voxel_spatial_size_prod_shift, dtype=point_voxel_xyz.dtype
    )
    # [B, N]
    point_voxel_id = ops.sum(point_voxel_xyz_multiplied, axis=-1)

    if batch_size == 1:
        return point_voxel_id

    batch_multiplier = (
        ops.arange(batch_size, dtype="int32") * voxel_spatial_size_prod[0]
    )
    batch_multiplier = ops.cast(
        ops.expand_dims(batch_multiplier, axis=-1), point_voxel_id.dtype
    )
    return point_voxel_id + batch_multiplier


class PointToVoxel(keras.layers.Layer):
    """Voxelization layer."""

    def __init__(self, voxel_size, spatial_size, **kwargs):
        """Voxelization layer constructor.

        Args:
          voxel_size: voxel size in each xyz dimension.
          spatial_size: max/min range in each dim in global coordinate frame.
          name: layer name
          **kwargs: additional key value args (e.g. dtype) passed to the parent
            class.
        """
        super().__init__(**kwargs)
        dim = len(voxel_size)
        assert len(spatial_size) == 2 * dim, f"{spatial_size}"

        self._voxel_size = voxel_size
        self._spatial_size = spatial_size

        self._voxel_spatial_size = voxel_utils.compute_voxel_spatial_size(
            spatial_size, self._voxel_size
        )

    # TODO(tanzhenyu): consider using keras masking.
    def call(self, point_xyz, point_mask):
        """Dynamically voxelizes points.

        B: batch_size.
        N: number of points.
        dim: the input dimension.

        Args:
          point_xyz: [B, N, dim] point xyz in global coordinate relative to sdc.
          point_mask: [B, N] valid point mask.

        Returns:
          point_voxel_feature: [B, N, dim] voxel feature (delta_{x,y,z}).
          point_voxel_id: [B, N] voxel ID of each point. Invalid voxels have
            Id's set to 0.
          point_voxel_mask: [B, N] validpoint voxel boolean mask.
        """
        # [B, N, dim]
        # convert from point coordinate to voxel index
        point_voxel_xyz_float = ops.floor(
            point_xyz / ops.array(self._voxel_size, point_xyz.dtype) + 0.5
        )
        # [B, N, dim]
        # delta to the nearest voxel
        point_voxel_feature = ops.cast(
            point_xyz
            - (
                point_voxel_xyz_float
                * ops.array(self._voxel_size, dtype=point_voxel_xyz_float.dtype)
            ),
            point_xyz.dtype,
        )

        # [B, N, dim]
        point_voxel_xyz_int = ops.cast(point_voxel_xyz_float, "int32")
        # [dim]
        # get xmin, ymin, zmin
        voxel_origin = voxel_utils.compute_voxel_origin(
            self._spatial_size, self._voxel_size
        )

        # [B, N, dim]
        # convert point voxel to positive voxel index
        point_voxel_xyz = point_voxel_xyz_int - ops.cast(
            ops.expand_dims(ops.expand_dims(voxel_origin, axis=0), axis=0),
            point_voxel_xyz_int.dtype,
        )

        # [B, N]
        # remove points outside the voxel boundary
        point_voxel_mask = ops.logical_and(
            point_voxel_xyz >= 0,
            point_voxel_xyz
            < ops.array(self._voxel_spatial_size, dtype=point_voxel_xyz.dtype),
        )
        point_voxel_mask = ops.all(point_voxel_mask, axis=-1)
        point_voxel_mask = ops.logical_and(point_voxel_mask, point_mask)

        # [B, N]
        point_voxel_mask_int = ops.cast(point_voxel_mask, dtype="int32")
        # [B, N] for voxel_id, int constant for num_voxels, in the range of
        # [0, B * num_voxels]
        point_voxel_id = ops.cast(
            compute_point_voxel_id(point_voxel_xyz, self._voxel_spatial_size),
            point_voxel_mask_int.dtype,
        )
        # [B, N]
        point_voxel_id = point_voxel_id * point_voxel_mask_int

        return point_voxel_feature, point_voxel_id, point_voxel_mask


@keras_cv_export("keras_cv.layers.DynamicVoxelization")
class DynamicVoxelization(keras.layers.Layer):
    """Dynamic voxelization and pool layer.

    This layer assigns and pools points into voxels,
    then it concatenates with point features and feed into a neural network,
    and max pools all point features inside each voxel.

    Args:
      voxel_size: the x, y, z dimension of each voxel.
      spatial_size: the x, y, z boundary of voxels

    Returns:
      voxelized feature, a float Tensor.

    """

    def __init__(self, voxel_size, spatial_size, **kwargs):
        super().__init__(**kwargs)
        self._voxelization_layer = PointToVoxel(
            voxel_size=voxel_size, spatial_size=spatial_size
        )
        self._voxel_size = voxel_size
        self._spatial_size = spatial_size
        self._voxel_spatial_size = voxel_utils.compute_voxel_spatial_size(
            spatial_size, self._voxel_size
        )
        self._voxel_spatial_size_volume = np.prod(
            self._voxel_spatial_size
        ).item()
        self.point_net_dense = keras.layers.Dense(128)
        self.point_net_norm = keras.layers.BatchNormalization()
        self.point_net_activation = keras.layers.ReLU()
        self.built = True

    def call(self, point_xyz, point_feature, point_mask, training=True):
        """Voxelizes and learns voxel features with a point net.

        B: batch_size.
        N: number of points.
        dim: the input dimension.

        Args:
          point_xyz: [B, N, 3] point xyz in global coordinate.
          point_feature: [B, N, dim] point feature inputs.
          point_mask: [B, N] valid point mask.
          training: whether it is in training mode.

        Returns:
          voxel_feature: [B, x_max, y_max, {z_max,}, mlp_dimension] voxel
            features. If z_max is 1, z-dim is squeezed.
        """
        (
            point_voxel_feature,
            point_voxel_id,
            point_voxel_mask,
        ) = self._voxelization_layer(point_xyz=point_xyz, point_mask=point_mask)
        # TODO(tanzhenyu): move compute_point_voxel_id to here, so PointToVoxel
        #  layer is more generic.
        point_feature = ops.concatenate(
            [point_feature, point_voxel_feature], axis=-1
        )
        batch_size = list(point_feature.shape)[0] or ops.shape(point_feature)[0]
        # [B, N, 1]
        point_mask_float = ops.expand_dims(
            ops.cast(point_voxel_mask, point_feature.dtype), axis=-1
        )
        # [B, N, dim]
        point_feature = point_feature * point_mask_float
        point_feature = self.point_net_dense(point_feature)
        point_feature = self.point_net_norm(
            point_feature, training=training, mask=point_mask
        )
        point_feature = self.point_net_activation(point_feature)
        # [B, N, new_dim]
        point_feature = point_feature * point_mask_float
        new_dim = list(point_feature.shape)[-1]
        point_feature = ops.reshape(point_feature, [-1, new_dim])
        point_voxel_id = ops.cast(ops.reshape(point_voxel_id, [-1]), "int32")
        # [B * num_voxels, new_dim]
        voxel_feature = ops.segment_max(
            point_feature,
            point_voxel_id,
            batch_size * self._voxel_spatial_size_volume,
        )
        # unsorted_segment_max sets empty values to -inf(float).
        voxel_feature_valid_mask = voxel_feature > VOXEL_FEATURE_MIN
        voxel_feature = voxel_feature * ops.cast(
            voxel_feature_valid_mask, dtype=voxel_feature.dtype
        )
        out_shape = [batch_size] + self._voxel_spatial_size + [new_dim]
        if out_shape[-2] == 1:
            out_shape = out_shape[:-2] + [out_shape[-1]]
        voxel_feature = ops.reshape(voxel_feature, out_shape)
        return voxel_feature

    def compute_output_shape(self, input_shape):
        return tuple([input_shape[0]] + self._voxel_spatial_size[:-1] + [128])



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ConvNeXt models for Keras.
References:
- [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545)
  (CVPR 2022)
"""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend
from tensorflow.keras import layers

from keras_cv.src.layers.regularization import StochasticDepth
from keras_cv.src.models.legacy import utils

MODEL_CONFIGS = {
    "tiny": {
        "depths": [3, 3, 9, 3],
        "projection_dims": [96, 192, 384, 768],
        "default_size": 224,
    },
    "small": {
        "depths": [3, 3, 27, 3],
        "projection_dims": [96, 192, 384, 768],
        "default_size": 224,
    },
    "base": {
        "depths": [3, 3, 27, 3],
        "projection_dims": [128, 256, 512, 1024],
        "default_size": 224,
    },
    "large": {
        "depths": [3, 3, 27, 3],
        "projection_dims": [192, 384, 768, 1536],
        "default_size": 224,
    },
    "xlarge": {
        "depths": [3, 3, 27, 3],
        "projection_dims": [256, 512, 1024, 2048],
        "default_size": 224,
    },
}

BASE_DOCSTRING = """Instantiates the {name} architecture.
    - [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545) (CVPR 2022)

    This function returns a Keras {name} model.
    Args:
        include_rescaling: bool, whether to rescale the inputs. If set
            to `True`, inputs will be passed through a `Rescaling(1/255.0)`
            layer.
        include_top: bool, whether to include the fully-connected layer at the
            top of the network. If provided, `num_classes` must be provided.
        depths: an iterable containing depths for each individual stages.
        projection_dims: An iterable containing output number of channels of
            each individual stages.
        drop_path_rate: stochastic depth probability, if 0.0, then stochastic
            depth won't be used.
        layer_scale_init_value: layer scale coefficient, if 0.0, layer scaling
            won't be used.
        weights: one of `None` (random initialization), a pretrained weight file
            path, or a reference to pre-trained weights (e.g.
            'imagenet/classification')(see available pre-trained weights in
            weights.py)
        input_shape: optional shape tuple, defaults to (None, None, 3).
        input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be the 4D tensor
                output of the last convolutional block.
            - `avg` means that global average pooling will be applied to the
                output of the last convolutional block, and thus the output of
                the model will be a 2D tensor.
            - `max` means that global max pooling will be applied.
        num_classes: optional int, number of classes to classify images into
            (only to be specified if `include_top` is `True`).
        classifier_activation: A `str` or callable. The activation function to
            use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top"
            layer.
        name: (Optional) name to pass to the model, defaults to "{name}".

    Returns:
      A `keras.Model` instance.
"""


@keras.utils.register_keras_serializable(package="keras_cv")
class LayerScale(layers.Layer):
    """Layer scale module.
    References:
      - https://arxiv.org/abs/2103.17239
    Args:
      init_values (float): Initial value for layer scale. Should be within
        [0, 1].
      projection_dim (int): Projection dimensionality.
    Returns:
      Tensor multiplied to the scale.
    """

    def __init__(self, init_values, projection_dim, **kwargs):
        super().__init__(**kwargs)
        self.init_values = init_values
        self.projection_dim = projection_dim

    def build(self, input_shape):
        self.gamma = tf.Variable(
            self.init_values * tf.ones((self.projection_dim,))
        )

    def call(self, x):
        return x * self.gamma

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "init_values": self.init_values,
                "projection_dim": self.projection_dim,
            }
        )
        return config


def apply_block(
    x,
    projection_dim,
    drop_path_rate=0.0,
    layer_scale_init_value=1e-6,
    name=None,
):
    """ConvNeXt block.
    References:
      - https://arxiv.org/abs/2201.03545
      - https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py
    Notes:
      In the original ConvNeXt implementation (linked above), the authors use
      `Dense` layers for pointwise convolutions for increased efficiency.
      Following that, this implementation also uses the same.
    Args:
      projection_dim (int): Number of filters for convolution layers. In the
        ConvNeXt paper, this is referred to as projection dimension.
      drop_path_rate (float): Probability of dropping paths. Should be within
        [0, 1].
      layer_scale_init_value (float): Layer scale value. Should be a small float
        number.
      name: name to path to the keras layer.
    Returns:
      A function representing a ConvNeXtBlock block.
    """  # noqa: E501
    if name is None:
        name = "prestem" + str(backend.get_uid("prestem"))

    inputs = x

    x = layers.Conv2D(
        filters=projection_dim,
        kernel_size=7,
        padding="same",
        groups=projection_dim,
        name=name + "_depthwise_conv",
    )(x)
    x = layers.LayerNormalization(epsilon=1e-6, name=name + "_layernorm")(x)
    x = layers.Dense(4 * projection_dim, name=name + "_pointwise_conv_1")(x)
    x = layers.Activation("gelu", name=name + "_gelu")(x)
    x = layers.Dense(projection_dim, name=name + "_pointwise_conv_2")(x)

    if layer_scale_init_value is not None:
        x = LayerScale(
            layer_scale_init_value,
            projection_dim,
            name=name + "_layer_scale",
        )(x)
    if drop_path_rate:
        layer = StochasticDepth(drop_path_rate, name=name + "_stochastic_depth")
        return layer([inputs, x])
    else:
        layer = layers.Activation("linear", name=name + "_identity")
        return inputs + layer(x)


def apply_head(x, num_classes, activation="softmax", name=None):
    """Implementation of classification head of ConvNeXt.
    Args:
      num_classes: number of classes for Dense layer
      activation: activation function for Dense layer
      name: name prefix
    Returns:
      Classification head function.
    """
    if name is None:
        name = str(backend.get_uid("head"))

    x = layers.GlobalAveragePooling2D(name=name + "_head_gap")(x)
    x = layers.LayerNormalization(epsilon=1e-6, name=name + "_head_layernorm")(
        x
    )
    x = layers.Dense(
        num_classes, activation=activation, name=name + "_head_dense"
    )(x)
    return x


@keras.utils.register_keras_serializable(package="keras_cv.models")
class ConvNeXt(keras.Model):
    """Instantiates ConvNeXt architecture given specific configuration.
    Args:
        include_rescaling: bool, whether to rescale the inputs. If set
            to `True`, inputs will be passed through a `Rescaling(1/255.0)`
            layer.
        include_top: bool, whether to include the fully-connected layer at the
            top of the network. If provided, `num_classes` must be provided.
        depths: An iterable containing depths for each individual stages.
        projection_dims: An iterable containing output number of channels of
        each individual stages.
        drop_path_rate: Stochastic depth probability. If 0.0, then stochastic
        depth won't be used.
        layer_scale_init_value: Layer scale coefficient. If 0.0, layer scaling
        won't be used.
        weights: one of `None` (random initialization), a pretrained weight file
            path, or a reference to pre-trained weights (e.g.
            'imagenet/classification')(see available pre-trained weights in
            weights.py)
        input_shape: optional shape tuple, defaults to (None, None, 3).
        input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be the 4D tensor
                output of the last convolutional block.
            - `avg` means that global average pooling will be applied to the
                output of the last convolutional block, and thus the output of
                the model will be a 2D tensor.
            - `max` means that global max pooling will be applied.
        num_classes: optional int, number of classes to classify images into
            (only to be specified if `include_top` is `True`).
        classifier_activation: A `str` or callable. The activation function to
            use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top"
            layer.
        name: (Optional) name to pass to the model, defaults to "convnext".
    Returns:
      A `keras.Model` instance.
    Raises:
        ValueError: in case of invalid argument for `weights`, or invalid input
            shape.
        ValueError: if `classifier_activation` is not `softmax`, or `None` when
            using a pretrained top layer.
        ValueError: if `include_top` is True but `num_classes` is not specified.
    """

    def __init__(
        self,
        include_rescaling,
        include_top,
        depths,
        projection_dims,
        drop_path_rate=0.0,
        layer_scale_init_value=1e-6,
        weights=None,
        input_shape=(None, None, 3),
        input_tensor=None,
        pooling=None,
        num_classes=None,
        classifier_activation="softmax",
        name="convnext",
        **kwargs,
    ):
        if weights and not tf.io.gfile.exists(weights):
            raise ValueError(
                "The `weights` argument should be either "
                "`None` or the path to the weights file to be loaded. "
                f"Weights file not found at location: {weights}"
            )

        if include_top and not num_classes:
            raise ValueError(
                "If `include_top` is True, "
                "you should specify `num_classes`. "
                f"Received: num_classes={num_classes}"
            )

        if include_top and pooling:
            raise ValueError(
                f"`pooling` must be `None` when `include_top=True`."
                f"Received pooling={pooling} and include_top={include_top}. "
            )

        inputs = utils.parse_model_inputs(input_shape, input_tensor)

        x = inputs
        if include_rescaling:
            x = layers.Rescaling(1 / 255.0)(x)

        # Stem block.
        stem = keras.Sequential(
            [
                layers.Conv2D(
                    projection_dims[0],
                    kernel_size=4,
                    strides=4,
                    name=name + "_stem_conv",
                ),
                layers.LayerNormalization(
                    epsilon=1e-6, name=name + "_stem_layernorm"
                ),
            ],
            name=name + "_stem",
        )

        # Downsampling blocks.
        downsample_layers = []
        downsample_layers.append(stem)

        num_downsample_layers = 3
        for i in range(num_downsample_layers):
            downsample_layer = keras.Sequential(
                [
                    layers.LayerNormalization(
                        epsilon=1e-6,
                        name=name + "_downsampling_layernorm_" + str(i),
                    ),
                    layers.Conv2D(
                        projection_dims[i + 1],
                        kernel_size=2,
                        strides=2,
                        name=name + "_downsampling_conv_" + str(i),
                    ),
                ],
                name=name + "_downsampling_block_" + str(i),
            )
            downsample_layers.append(downsample_layer)

        # Stochastic depth schedule.
        # This is referred from the original ConvNeXt codebase:
        # https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py#L86
        depth_drop_rates = [
            float(x) for x in tf.linspace(0.0, drop_path_rate, sum(depths))
        ]

        # First apply downsampling blocks and then apply ConvNeXt stages.
        cur = 0

        num_convnext_blocks = 4
        for i in range(num_convnext_blocks):
            x = downsample_layers[i](x)
            for j in range(depths[i]):
                x = apply_block(
                    x,
                    projection_dim=projection_dims[i],
                    drop_path_rate=depth_drop_rates[cur + j],
                    layer_scale_init_value=layer_scale_init_value,
                    name=name + f"_stage_{i}_block_{j}",
                )
            cur += depths[i]

        if include_top:
            x = apply_head(
                x,
                num_classes=num_classes,
                activation=classifier_activation,
                name=name,
            )

        else:
            if pooling == "avg":
                x = layers.GlobalAveragePooling2D()(x)
            elif pooling == "max":
                x = layers.GlobalMaxPooling2D()(x)
            x = layers.LayerNormalization(epsilon=1e-6)(x)

        # Create model.
        super().__init__(inputs=inputs, outputs=x, **kwargs)

        if weights is not None:
            self.load_weights(weights)

        self.include_rescaling = include_rescaling
        self.include_top = include_top
        self.depths = depths
        self.projection_dims = projection_dims
        self.drop_path_rate = drop_path_rate
        self.layer_scale_init_value = layer_scale_init_value
        self.input_tensor = input_tensor
        self.pooling = pooling
        self.num_classes = num_classes
        self.classifier_activation = classifier_activation

    def get_config(self):
        return {
            "include_rescaling": self.include_rescaling,
            "include_top": self.include_top,
            "depths": self.depths,
            "projection_dims": self.projection_dims,
            "drop_path_rate": self.drop_path_rate,
            "layer_scale_init_value": self.layer_scale_init_value,
            # Remove batch dimension from `input_shape`
            "input_shape": self.input_shape[1:],
            "input_tensor": self.input_tensor,
            "pooling": self.pooling,
            "num_classes": self.num_classes,
            "classifier_activation": self.classifier_activation,
            "name": self.name,
            "trainable": self.trainable,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def ConvNeXtTiny(
    *,
    include_rescaling,
    include_top,
    drop_path_rate,
    layer_scale_init_value,
    weights=None,
    input_shape=(None, None, 3),
    input_tensor=None,
    pooling=None,
    num_classes=None,
    classifier_activation="softmax",
    name="convnext_tiny",
):
    return ConvNeXt(
        include_rescaling=include_rescaling,
        include_top=include_top,
        depths=MODEL_CONFIGS["tiny"]["depths"],
        projection_dims=MODEL_CONFIGS["tiny"]["projection_dims"],
        drop_path_rate=drop_path_rate,
        layer_scale_init_value=layer_scale_init_value,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        name=name,
    )


def ConvNeXtSmall(
    *,
    include_rescaling,
    include_top,
    drop_path_rate,
    layer_scale_init_value,
    weights=None,
    input_shape=(None, None, 3),
    input_tensor=None,
    pooling=None,
    num_classes=None,
    classifier_activation="softmax",
    name="convnext_small",
):
    return ConvNeXt(
        include_rescaling=include_rescaling,
        include_top=include_top,
        depths=MODEL_CONFIGS["small"]["depths"],
        projection_dims=MODEL_CONFIGS["small"]["projection_dims"],
        drop_path_rate=drop_path_rate,
        layer_scale_init_value=layer_scale_init_value,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        name=name,
    )


def ConvNeXtBase(
    *,
    include_rescaling,
    include_top,
    drop_path_rate,
    layer_scale_init_value,
    weights=None,
    input_shape=(None, None, 3),
    input_tensor=None,
    pooling=None,
    num_classes=None,
    classifier_activation="softmax",
    name="convnext_base",
):
    return ConvNeXt(
        include_rescaling=include_rescaling,
        include_top=include_top,
        depths=MODEL_CONFIGS["base"]["depths"],
        projection_dims=MODEL_CONFIGS["base"]["projection_dims"],
        drop_path_rate=drop_path_rate,
        layer_scale_init_value=layer_scale_init_value,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        name=name,
    )


def ConvNeXtLarge(
    *,
    include_rescaling,
    include_top,
    drop_path_rate,
    layer_scale_init_value,
    weights=None,
    input_shape=(None, None, 3),
    input_tensor=None,
    pooling=None,
    num_classes=None,
    classifier_activation="softmax",
    name="convnext_large",
):
    return ConvNeXt(
        include_rescaling=include_rescaling,
        include_top=include_top,
        depths=MODEL_CONFIGS["large"]["depths"],
        projection_dims=MODEL_CONFIGS["large"]["projection_dims"],
        drop_path_rate=drop_path_rate,
        layer_scale_init_value=layer_scale_init_value,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        name=name,
    )


def ConvNeXtXLarge(
    *,
    include_rescaling,
    include_top,
    drop_path_rate,
    layer_scale_init_value,
    weights=None,
    input_shape=(None, None, 3),
    input_tensor=None,
    pooling=None,
    num_classes=None,
    classifier_activation="softmax",
    name="convnext_xlarge",
):
    return ConvNeXt(
        include_rescaling=include_rescaling,
        include_top=include_top,
        depths=MODEL_CONFIGS["xlarge"]["depths"],
        projection_dims=MODEL_CONFIGS["xlarge"]["projection_dims"],
        drop_path_rate=drop_path_rate,
        layer_scale_init_value=layer_scale_init_value,
        weights=weights,
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        name=name,
    )


ConvNeXtTiny.__doc__ = BASE_DOCSTRING.format(name="ConvNeXtTiny")
ConvNeXtSmall.__doc__ = BASE_DOCSTRING.format(name="ConvNeXtSmall")
ConvNeXtBase.__doc__ = BASE_DOCSTRING.format(name="ConvNeXtBase")
ConvNeXtLarge.__doc__ = BASE_DOCSTRING.format(name="ConvNeXtLarge")
ConvNeXtXLarge.__doc__ = BASE_DOCSTRING.format(name="ConvNeXtXLarge")



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import tensorflow as tf
from tensorflow.keras import layers

from keras_cv.src.api_export import keras_cv_export


@keras_cv_export("keras_cv.layers.PatchingAndEmbedding")
class PatchingAndEmbedding(layers.Layer):
    """
    Layer to patchify images, prepend a class token, positionally embed and
    create a projection of patches for Vision Transformers

    The layer expects a batch of input images and returns batches of patches,
    flattened as a sequence and projected onto `project_dims`. If the height and
    width of the images aren't divisible by the patch size, the supplied padding
    type is used (or 'VALID' by default).

    Reference:
        An Image is Worth 16x16 Words: Transformers for Image Recognition at
        Scale by Alexey Dosovitskiy et al. (https://arxiv.org/abs/2010.11929)

    Args:
        project_dim: the dimensionality of the project_dim
        patch_size: the patch size
        padding: default 'VALID', the padding to apply for patchifying images

    Returns:
        Patchified and linearly projected input images, including a prepended
        learnable class token with shape (batch, num_patches+1, project_dim)

    Example:

    ```
    images = #... batch of images
    encoded_patches = keras_cv.layers.PatchingAndEmbedding(
        project_dim=project_dim,
        patch_size=patch_size)(patches)
    print(encoded_patches.shape) # (1, 197, 1024)
    ```
    """

    def __init__(self, project_dim, patch_size, padding="VALID", **kwargs):
        super().__init__(**kwargs)
        self.project_dim = project_dim
        self.patch_size = patch_size
        self.padding = padding
        if patch_size < 0:
            raise ValueError(
                "The patch_size cannot be a negative number. Received "
                f"{patch_size}"
            )
        if padding not in ["VALID", "SAME"]:
            raise ValueError(
                f"Padding must be either 'SAME' or 'VALID', but {padding} was "
                "passed."
            )
        self.projection = layers.Conv2D(
            filters=self.project_dim,
            kernel_size=self.patch_size,
            strides=self.patch_size,
            padding=self.padding,
        )

    def build(self, input_shape):
        self.class_token = self.add_weight(
            shape=[1, 1, self.project_dim], name="class_token", trainable=True
        )
        self.num_patches = (
            input_shape[1]
            // self.patch_size
            * input_shape[2]
            // self.patch_size
        )
        self.position_embedding = layers.Embedding(
            input_dim=self.num_patches + 1, output_dim=self.project_dim
        )

    def call(
        self,
        images,
        interpolate=False,
        interpolate_width=None,
        interpolate_height=None,
        patch_size=None,
    ):
        """Calls the PatchingAndEmbedding layer on a batch of images.
        Args:
            images: A `tf.Tensor` of shape [batch, width, height, depth]
            interpolate: A `bool` to enable or disable interpolation
            interpolate_height: An `int` representing interpolated height
            interpolate_width: An `int` representing interpolated width
            patch_size: An `int` representing the new patch size if
                interpolation is used

        Returns:
            `A tf.Tensor` of shape [batch, patch_num+1, embedding_dim]
        """
        # Turn images into patches and project them onto `project_dim`
        patches = self.projection(images)
        patch_shapes = tf.shape(patches)
        patches_flattened = tf.reshape(
            patches,
            shape=(
                patch_shapes[0],
                patch_shapes[-2] * patch_shapes[-2],
                patch_shapes[-1],
            ),
        )

        # Add learnable class token before linear projection and positional
        # embedding
        flattened_shapes = tf.shape(patches_flattened)
        class_token_broadcast = tf.cast(
            tf.broadcast_to(
                self.class_token,
                [flattened_shapes[0], 1, flattened_shapes[-1]],
            ),
            dtype=patches_flattened.dtype,
        )
        patches_flattened = tf.concat(
            [class_token_broadcast, patches_flattened], 1
        )
        positions = tf.range(start=0, limit=self.num_patches + 1, delta=1)

        if interpolate and None not in (
            interpolate_width,
            interpolate_height,
            patch_size,
        ):
            (
                interpolated_embeddings,
                class_token,
            ) = self.__interpolate_positional_embeddings(
                self.position_embedding(positions),
                interpolate_width,
                interpolate_height,
                patch_size,
            )
            addition = patches_flattened + interpolated_embeddings
            encoded = tf.concat([class_token, addition], 1)
        elif interpolate and None in (
            interpolate_width,
            interpolate_height,
            patch_size,
        ):
            raise ValueError(
                "`None of `interpolate_width`, `interpolate_height` and "
                "`patch_size` cannot be None if `interpolate` is True"
            )
        else:
            encoded = patches_flattened + self.position_embedding(positions)
        return encoded

    def __interpolate_positional_embeddings(
        self, embedding, height, width, patch_size
    ):
        """
        Allows for pre-trained position embedding interpolation. This trick
        allows you to fine-tune a ViT on higher resolution images than it was
        trained on.

        Based on:
        https://github.com/huggingface/transformers/blob/main/src/transformers/models/vit/modeling_tf_vit.py
        """

        dimensionality = embedding.shape[-1]

        class_token = tf.expand_dims(embedding[:1, :], 0)
        patch_positional_embeddings = embedding[1:, :]

        h0 = height // patch_size
        w0 = width // patch_size

        new_shape = tf.constant(int(math.sqrt(self.num_patches)))

        interpolated_embeddings = tf.image.resize(
            images=tf.reshape(
                patch_positional_embeddings,
                shape=(
                    1,
                    new_shape,
                    new_shape,
                    dimensionality,
                ),
            ),
            size=(h0, w0),
            method="bicubic",
        )

        reshaped_embeddings = tf.reshape(
            tensor=interpolated_embeddings, shape=(1, -1, dimensionality)
        )

        # linear_projection = self.linear_projection(reshaped_embeddings)
        # addition = linear_projection + reshaped_embeddings

        # return tf.concat([class_token, addition], 1)
        return reshaped_embeddings, class_token

    def get_config(self):
        config = {
            "project_dim": self.project_dim,
            "patch_size": self.patch_size,
            "padding": self.padding,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.models.classification.image_classifier import ImageClassifier
from keras_cv.src.models.classification.video_classifier import VideoClassifier



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models.segmentation.segment_anything.sam_layers import (
    RandomFrequencyPositionalEmbeddings,
)


@keras_cv_export("keras_cv.models.SAMPromptEncoder", package="keras_cv.models")
class SAMPromptEncoder(keras.layers.Layer):
    """Prompt Encoder for the Segment Anything Model (SAM).

    The prompt encoder generates encodings for three types of prompts:

    - Point prompts: Points on the image along with a label indicating whether
        the point is in the foreground (part of the mask) or in the background
        (not a part of the mask).
    - Box prompts: A batch of bounding boxes with format [(x1, y1), (x2, y2)]
        used to determine the location of the masks in the image.
    - Masks: An input mask can be passed to refine the positional embeddings
        for the output mask.

    First, the point prompts and box prompts are concatenated and positional
    encodings are generated using random spatial frequencies. A point is
    represented as the sum of a positional encoding of the point's location
    and one of two learned embeddings that indicate if the point is either in
    the foreground or background. A box is represented by an embedding pair:

    (1) the positional encoding of its top-left corner summed with a learned
    embedding representing "top-left corner" and
    (2) the same structure but using a learned embedding indicating
    "bottom-right corner".

    The box and point encodings are referred to as "sparse encodings"

    If a mask prompt is passed, a convolutional neural net is used to
    downscale it to generate "dense encodings". If no mask prompt is passed,
    an embedding layer is used instead to generate a "no mask" embedding.

    Args:
        embed_dim (int, optional): The number of features in the output
            embeddings. Defaults to `256`.
        image_embedding_size (int, optional): The number of features in the
            image embeddings generated by an image encoder. Defaults to
            `(64, 64)`.
        input_image_size (tuple[int], optional): A tuple of the height and
            width of the image being prompted. Defaults to `(1024, 1024)`.
        mask_in_chans (int, optional): The number of channels of the mask
            prompt. Defaults to `16`.
        activation (str, optional): The activation to use in the mask
            downscaler neural net. Defaults to `"gelu"`.

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)
    """  # noqa: E501

    def __init__(
        self,
        *,
        embed_dim=256,
        image_embedding_size=(64, 64),
        input_image_size=(1024, 1024),
        mask_in_chans=16,
        activation="gelu",
        **kwargs
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.image_embedding_size = image_embedding_size
        self.input_image_size = input_image_size
        self.mask_in_chans = mask_in_chans
        self.activation = activation

        self.positional_embedding_layer = RandomFrequencyPositionalEmbeddings(
            num_positional_features=self.embed_dim // 2, scale=1
        )

        self.foreground_point_embed = keras.layers.Embedding(
            1, embed_dim, name="foreground_point_embed"
        )
        self.background_point_embed = keras.layers.Embedding(
            1, embed_dim, name="background_point_embed"
        )
        self.top_left_corner_embed = keras.layers.Embedding(
            1, embed_dim, name="top_left_corner_embed"
        )
        self.bottom_right_corner_embed = keras.layers.Embedding(
            1, embed_dim, name="bottom_right_corner_embed"
        )
        self.not_a_point_embed = keras.layers.Embedding(
            1, embed_dim, name="not_a_point_embed"
        )

        self.mask_downscaler = keras.models.Sequential(
            [
                keras.layers.Conv2D(
                    mask_in_chans // 4, kernel_size=2, strides=2
                ),
                keras.layers.LayerNormalization(epsilon=1e-6),
                keras.layers.Activation(activation),
                keras.layers.Conv2D(mask_in_chans, kernel_size=2, strides=2),
                keras.layers.LayerNormalization(epsilon=1e-6),
                keras.layers.Activation(activation),
                keras.layers.Conv2D(embed_dim, kernel_size=1),
            ],
            name="mask_downscaler",
        )
        self.no_mask_embed = keras.layers.Embedding(
            1, embed_dim, name="no_mask_embed"
        )

    def build(self, input_shape=None):
        self.positional_embedding_layer.build()
        for layer in [
            self.foreground_point_embed,
            self.background_point_embed,
            self.top_left_corner_embed,
            self.bottom_right_corner_embed,
            self.not_a_point_embed,
            self.no_mask_embed,
        ]:
            layer.build([None])
        self.mask_downscaler.build(
            [
                None,
                4 * self.image_embedding_size[0],
                4 * self.image_embedding_size[1],
                1,
            ]
        )
        self.built = True

    def compute_output_shape(self, input_shape):
        return {
            "sparse_embeddings": [None, None, self.embed_dim],
            "dense_embeddings": [
                None,
                self.image_embedding_size[0],
                self.image_embedding_size[1],
                self.embed_dim,
            ],
            "dense_positional_embeddings": [
                None,
                self.image_embedding_size[0],
                self.image_embedding_size[1],
                self.embed_dim,
            ],
        }

    def __embed_points(self, points, labels):
        points = points + 0.5
        indices = ops.arange(1, dtype="int32")

        point_embeddings = self.positional_embedding_layer.encode_coordinates(
            points, self.input_image_size
        )
        labels = ops.broadcast_to(
            labels[..., None], ops.shape(point_embeddings)
        )
        point_embeddings = ops.where(
            labels == 0,
            point_embeddings + self.background_point_embed(indices),
            point_embeddings + self.foreground_point_embed(indices),
        )
        point_embeddings = ops.where(
            labels == -1,
            self.not_a_point_embed(indices),
            point_embeddings,
        )
        return point_embeddings

    def __embed_box(self, box):
        shape = ops.shape(box)
        B, N = shape[0], shape[1]
        box = box + 0.5
        indices = ops.arange(1, dtype="int32")
        corner_embedding = self.positional_embedding_layer.encode_coordinates(
            box, self.input_image_size
        )
        top_left_embedding = corner_embedding[
            :, :, 0, :
        ] + self.top_left_corner_embed(indices)
        bottom_right_embedding = corner_embedding[
            :, :, 1, :
        ] + self.bottom_right_corner_embed(indices)
        corner_embedding = ops.stack(
            [top_left_embedding, bottom_right_embedding], axis=2
        )
        return ops.reshape(corner_embedding, (B, N * 2, self.embed_dim))

    def __embed_mask(self, mask):
        mask_embedding = self.mask_downscaler(mask)
        return mask_embedding

    def call(self, inputs):
        # Get the batch shape based on any arbitrary input, because batch
        # shapes must all match.
        B = ops.shape(next(iter(inputs.values())))[0]

        points = inputs.get("points", ops.zeros((B, 0, 2)))
        labels = inputs.get("labels", ops.zeros((B, 0)))
        box = inputs.get("boxes", ops.zeros((B, 0, 2, 2)))
        mask = inputs.get("masks", ops.zeros((B, 0, 256, 256, 1)))

        # Compute point embeddings
        point_embeddings = self.__embed_points(points, labels)

        # Compute box embeddings
        box_embeddings = self.__embed_box(box)

        # Concatenate both into a sparse embeddings tensor
        sparse_embeddings = ops.concatenate(
            [point_embeddings, box_embeddings], axis=1
        )

        # Compute the mask embeddings
        _no_mask_embed = lambda: (
            ops.broadcast_to(
                ops.reshape(
                    self.no_mask_embed(ops.arange(1, dtype="int32")),
                    (1, 1, 1, self.embed_dim),
                ),
                shape=(
                    B,
                    self.image_embedding_size[0],
                    self.image_embedding_size[1],
                    self.embed_dim,
                ),
            )
        )

        def _maybe_input_mask_embed():
            # Keras Core passes the masks as concrete tensors for both the
            # true and false functions to build the output shape. So, we
            # need to handle the case when 0 size mask is passed and
            # dispatch the call to `_no_mask_embed`. Note that we can't call
            # the lambda directly since the inputs are bound to different
            # values when called with concrete values.
            if mask.shape[1] == 0:
                return ops.broadcast_to(
                    ops.reshape(
                        self.no_mask_embed(ops.arange(1, dtype="int32")),
                        (1, 1, 1, self.embed_dim),
                    ),
                    shape=(
                        B,
                        self.image_embedding_size[0],
                        self.image_embedding_size[1],
                        self.embed_dim,
                    ),
                )
            shape = ops.shape(mask)
            BM, N, H, W, C = shape[0], shape[1], shape[2], shape[3], shape[4]
            return self.__embed_mask(ops.reshape(mask, (BM * N, H, W, C)))

        dense_embeddings = ops.cond(
            ops.equal(ops.size(mask), 0),
            _no_mask_embed,
            _maybe_input_mask_embed,
        )

        # Compute the dense positional embeddings
        dense_positional_embeddings = (
            self.positional_embedding_layer.encode_image(
                self.image_embedding_size
            )[None, ...]
        )

        return {
            "sparse_embeddings": sparse_embeddings,
            "dense_embeddings": dense_embeddings,
            "dense_positional_embeddings": dense_positional_embeddings,
        }

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "embed_dim": self.embed_dim,
                "image_embedding_size": self.image_embedding_size,
                "input_image_size": self.input_image_size,
                "mask_in_chans": self.mask_in_chans,
                "activation": self.activation,
            }
        )
        return config



# Copyright 2022 Waymo LLC.
#
# Licensed under the terms in https://github.com/keras-team/keras-cv/blob/master/keras_cv/layers/preprocessing_3d/waymo/LICENSE  # noqa: E501

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import random
from keras_cv.src.bounding_box_3d import CENTER_XYZ_DXDYDZ_PHI
from keras_cv.src.layers.preprocessing_3d import base_augmentation_layer_3d
from keras_cv.src.point_cloud import coordinate_transform
from keras_cv.src.point_cloud import wrap_angle_radians

POINT_CLOUDS = base_augmentation_layer_3d.POINT_CLOUDS
BOUNDING_BOXES = base_augmentation_layer_3d.BOUNDING_BOXES


@keras_cv_export("keras_cv.layers.GlobalRandomRotation")
class GlobalRandomRotation(base_augmentation_layer_3d.BaseAugmentationLayer3D):
    """A preprocessing layer which randomly rotates point clouds and bounding
    boxes along X, Y and Z axes during training.

    This layer will randomly rotate the whole scene along the X, Y and Z axes
    based on a randomly sampled rotation angle between [-max_rotation_angle,
    max_rotation_angle] (in radians) following a uniform distribution. During
    inference time, the output will be identical to input. Call the layer with
    `training=True` to rotate the input.

    Input shape:
      point_clouds: 3D (multi frames) float32 Tensor with shape
        [num of frames, num of points, num of point features].
        The first 5 features are [x, y, z, class, range].
      bounding_boxes: 3D (multi frames) float32 Tensor with shape
        [num of frames, num of boxes, num of box features]. Boxes are expected
        to follow the CENTER_XYZ_DXDYDZ_PHI format. Refer to
        https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box_3d/formats.py
        for more details on supported bounding box formats.

    Output shape:
      A dictionary of Tensors with the same shape as input Tensors.

    Arguments:
      max_rotation_angle_x: A float scalar sets the maximum rotation angle (in
        radians) along X axis.
      max_rotation_angle_y: A float scalar sets the maximum rotation angle (in
        radians) along Y axis.
      max_rotation_angle_z: A float scalar sets the maximum rotation angle (in
        radians) along Z axis.

    """

    def __init__(
        self,
        max_rotation_angle_x=None,
        max_rotation_angle_y=None,
        max_rotation_angle_z=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        max_rotation_angle_x = (
            max_rotation_angle_x if max_rotation_angle_x else 0.0
        )
        max_rotation_angle_y = (
            max_rotation_angle_y if max_rotation_angle_y else 0.0
        )
        max_rotation_angle_z = (
            max_rotation_angle_z if max_rotation_angle_z else 0.0
        )

        if max_rotation_angle_x < 0:
            raise ValueError("max_rotation_angle_x must be >=0.")
        if max_rotation_angle_y < 0:
            raise ValueError("max_rotation_angle_y must be >=0.")
        if max_rotation_angle_z < 0:
            raise ValueError("max_rotation_angle_z must be >=0.")
        self._max_rotation_angle_x = max_rotation_angle_x
        self._max_rotation_angle_y = max_rotation_angle_y
        self._max_rotation_angle_z = max_rotation_angle_z

    def get_config(self):
        return {
            "max_rotation_angle_x": self._max_rotation_angle_x,
            "max_rotation_angle_y": self._max_rotation_angle_y,
            "max_rotation_angle_z": self._max_rotation_angle_z,
        }

    def get_random_transformation(self, **kwargs):
        random_rotation_x = random.uniform(
            (),
            minval=-self._max_rotation_angle_x,
            maxval=self._max_rotation_angle_x,
            dtype=self.compute_dtype,
            seed=self._random_generator,
        )
        random_rotation_y = random.uniform(
            (),
            minval=-self._max_rotation_angle_y,
            maxval=self._max_rotation_angle_y,
            dtype=self.compute_dtype,
            seed=self._random_generator,
        )
        random_rotation_z = random.uniform(
            (),
            minval=-self._max_rotation_angle_z,
            maxval=self._max_rotation_angle_z,
            dtype=self.compute_dtype,
            seed=self._random_generator,
        )
        return {
            "pose": tf.stack(
                [
                    0,
                    0,
                    0,
                    random_rotation_z,
                    random_rotation_x,
                    random_rotation_y,
                ],
                axis=0,
            )
        }

    def augment_point_clouds_bounding_boxes(
        self, point_clouds, bounding_boxes, transformation, **kwargs
    ):
        pose = transformation["pose"]
        point_clouds_xyz = coordinate_transform(point_clouds[..., :3], pose)
        point_clouds = tf.concat(
            [point_clouds_xyz, point_clouds[..., 3:]], axis=-1
        )

        bounding_boxes_xyz = coordinate_transform(
            bounding_boxes[..., : CENTER_XYZ_DXDYDZ_PHI.Z + 1], pose
        )
        bounding_boxes_heading = wrap_angle_radians(
            tf.expand_dims(
                bounding_boxes[..., CENTER_XYZ_DXDYDZ_PHI.PHI], axis=-1
            )
            - pose[3]
        )
        bounding_boxes = tf.concat(
            [
                bounding_boxes_xyz,
                bounding_boxes[
                    ..., CENTER_XYZ_DXDYDZ_PHI.DX : CENTER_XYZ_DXDYDZ_PHI.DZ + 1
                ],
                bounding_boxes_heading,
                bounding_boxes[..., CENTER_XYZ_DXDYDZ_PHI.CLASS :],
            ],
            axis=-1,
        )

        return (point_clouds, bounding_boxes)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BASNet model preset configurations."""

from keras_cv.src.models.backbones.resnet_v1 import resnet_v1_backbone_presets

presets_no_weights = {
    "basnet_resnet18": {
        "metadata": {
            "description": "BASNet with a ResNet18 v1 backbone.",
            "params": 98780872,
            "official_name": "BASNet",
            "path": "basnet_resnet18",
        },
        "config": {
            "backbone": resnet_v1_backbone_presets.backbone_presets["resnet18"],
            "num_classes": 1,
            "input_shape": (288, 288, 3),
        },
    },
    "basnet_resnet34": {
        "metadata": {
            "description": "BASNet with a ResNet34 v1 backbone.",
            "params": 108896456,
            "official_name": "BASNet",
            "path": "basnet_resnet34",
        },
        "config": {
            "backbone": resnet_v1_backbone_presets.backbone_presets["resnet34"],
            "num_classes": 1,
            "input_shape": (288, 288, 3),
        },
    },
}

presets_with_weights = {
    # TODO: Add BASNet preset with weights
}

basnet_presets = {**presets_no_weights, **presets_with_weights}



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
import tensorflow as tf

from keras_cv.src.backend import ops

EPSILON = 1e-4


def compute_feature_map_ref_xyz(
    voxel_size,
    spatial_size,
    global_xyz,
):
    """Computes the offset xyz locations for each feature map pixel.

    Args:
      voxel_size: voxel size.
      spatial_size: the x, y, z boundary of voxels.
      global_xyz: [B, 3] tensor

    Returns:
      [B, H, W, Z, 3] offset locations for each feature map pixel in global
        coordinate.
    """
    voxel_spatial_size = compute_voxel_spatial_size(spatial_size, voxel_size)
    voxel_coord_meshgrid = np.mgrid[
        0 : voxel_spatial_size[0],
        0 : voxel_spatial_size[1],
        0 : voxel_spatial_size[2],
    ]
    voxel_coord = np.concatenate(voxel_coord_meshgrid[..., np.newaxis], axis=-1)
    # [H, W, Z, 3]
    # [3]
    voxel_origin = (compute_voxel_origin(spatial_size, voxel_size),)
    # [H, W, Z, 3]
    voxel_coord = voxel_coord + voxel_origin
    # [H, W, Z, 3]
    ref = ops.cast(voxel_coord * np.array(voxel_size), global_xyz.dtype)
    # [1, H, W, Z, 3] + [B, 1, 1, 1, 3] -> [B, H, W, Z, 3]
    ref = ops.expand_dims(ref, axis=0) + ops.expand_dims(
        ops.expand_dims(ops.expand_dims(global_xyz, axis=1), axis=1), axis=1
    )
    return ref


def compute_voxel_spatial_size(spatial_size, voxel_size):
    """Computes how many voxels in each dimension are needed.

    Args:
      spatial_size: max/min range in each dim in global coordinate frame.
      voxel_size: voxel size.

    Returns:
      voxel_spatial_size: voxel spatial size.
    """
    dim = len(voxel_size)
    # Compute the range as x_range = xmax - xmin, ymax - ymin, zmax - zmin
    voxel_spatial_size_float = [
        spatial_size[2 * i + 1] - spatial_size[2 * i] for i in range(dim)
    ]
    # voxel_dim_x / x_range
    voxel_spatial_size_float = [
        i / j for i, j in zip(voxel_spatial_size_float, voxel_size)
    ]
    voxel_spatial_size_int = [
        math.ceil(v - EPSILON) for v in voxel_spatial_size_float
    ]

    return voxel_spatial_size_int


def compute_voxel_origin(
    spatial_size,
    voxel_size,
):
    """Computes voxel origin.

    Args:
      spatial_size: The current location of SDC.
      voxel_size: 1.0 / voxel size.

    Returns:
      voxel_origin: [dim] the voxel origin.
    """
    voxel_origin = spatial_size[::2]
    voxel_origin = np.array(
        [o / v for o, v in zip(voxel_origin, voxel_size)], "float32"
    )
    voxel_origin = np.round(voxel_origin)
    return voxel_origin


def point_to_voxel_coord(point_xyz, voxel_size, dtype=tf.int32):
    """Computes the voxel coord given points.

    A voxel x represents [(x-0.5) / voxel_size, (x+0.5) / voxel_size)
    in the coordinate system of the input point_xyz.

    Args:
      point_xyz: [..., dim] point xyz coordinates.
      voxel_size: voxel size.
      dtype: the output dtype.

    Returns:
      voxelized coordinates.
    """
    with tf.name_scope("point_to_voxel_coord"):
        point_voxelized = point_xyz / tf.constant(
            voxel_size, dtype=point_xyz.dtype
        )
        assert dtype.is_integer or dtype.is_floating, f"{dtype}"
        # Note: tf.round casts float to the nearest integer. If the float is
        # 0.5, it casts it to the nearest even integer.
        point_voxelized_round = tf.math.round(point_voxelized)
        if dtype.is_floating:
            assert dtype == point_xyz.dtype, f"{dtype}"
            return point_voxelized_round
        return tf.cast(point_voxelized_round, dtype=dtype)


def voxel_coord_to_point(voxel_coord, voxel_size, dtype=tf.float32):
    """Convert voxel coord to expected point in the original coordinate system.

    This is the reverse of point_to_voxel_coord.

    Args:
      voxel_coord: [..., dim] int tensors for coordinate of each voxel.
      voxel_size: voxel size.
      dtype: output point data type.

    Returns:
      point coordinates.
    """
    with tf.name_scope("voxel_coord_to_point"):
        # This simply computes voxel_coord * voxel_size.
        if voxel_coord.dtype != dtype:
            voxel_coord = tf.cast(voxel_coord, dtype=dtype)
        return voxel_coord * tf.constant(voxel_size, dtype=dtype)


def get_yaw_rotation(yaw, name=None):
    """Gets a rotation matrix given yaw only.

    Args:
      yaw: x-rotation in radians. This tensor can be any shape except an empty
        one.
      name: the op name.

    Returns:
      A rotation tensor with the same data type of the input. Its shape is
        [input_shape, 3 ,3].
    """
    with tf.name_scope("GetYawRotation"):
        cos_yaw = tf.cos(yaw)
        sin_yaw = tf.sin(yaw)
        ones = tf.ones_like(yaw)
        zeros = tf.zeros_like(yaw)

        return tf.stack(
            [
                tf.stack([cos_yaw, -1.0 * sin_yaw, zeros], axis=-1),
                tf.stack([sin_yaw, cos_yaw, zeros], axis=-1),
                tf.stack([zeros, zeros, ones], axis=-1),
            ],
            axis=-2,
        )


def inv_loc(rot, loc):
    """Invert a location.
    rot and loc can form a transform matrix between two frames.
    R = rot, L = loc
    R*R' = I
    R * new_loc + L = 0 = > new_loc = -R'*L
    Args:
      rot: [..., 3, 3] rotation matrix.
      loc: [..., 3] location matrix.
    Returns:
      [..., 3] new location matrix.
    """
    new_loc = -1.0 * tf.linalg.matmul(
        rot, loc[..., tf.newaxis], transpose_a=True
    )
    return tf.squeeze(new_loc, axis=-1)


def _has_rank(tensor, expected_rank):
    """Syntactic sugar for asserting that tensor has the expected rank.

    Internal usages for keras_cv libraries only.
    """
    if tensor.shape.ndims is not None and isinstance(expected_rank, int):
        assert tensor.shape.ndims == expected_rank, (
            "Ranks did not match, got %d, " "expected %d"
        ) % (tensor.shape.ndims, expected_rank)
    return tensor


def _pad_or_trim_to(x, shape, pad_val=0, pad_after_contents=True):
    """Pad and slice x to the given shape.

    This is branched from Lingvo
    https://github.com/tensorflow/lingvo/blob/master/lingvo/core/py_utils.py.

    Internal usages for keras_cv libraries only.

    Args:
      x: A tensor.
      shape: The shape of the returned tensor.
      pad_val: An int or float used to pad x.
      pad_after_contents: Whether to pad and trim after the original contents of
        each dimension.
    Returns:
      'x' is padded with pad_val and sliced so that the result has the given
      shape.
    Raises:
      ValueError: if shape is a tf.TensorShape and not fully defined.
    """
    if isinstance(shape, (list, tuple)):
        expected_rank = len(shape)
    elif isinstance(shape, tf.TensorShape):
        if not shape.is_fully_defined():
            raise ValueError(
                "shape %s padding %s must be fully defined." % (shape, x)
            )
        expected_rank = shape.rank
    else:
        shape = _has_rank(shape, 1)
        expected_rank = tf.size(shape)
    x = _has_rank(x, expected_rank)

    pad = shape - tf.minimum(tf.shape(x), shape)
    zeros = tf.zeros_like(pad)
    if pad_after_contents:
        # If dim_i is less than shape[i], pads after contents.
        paddings = tf.stack([zeros, pad], axis=1)
        # If dim_i is larger than shape[i], we slice [0:shape[i]] for dim_i.
        slice_begin = zeros
    else:
        # If dim_i is less than shape[i], pads before contents.
        paddings = tf.stack([pad, zeros], axis=1)
        # If dim-i is larger than shape[i], we slice [dim_i - shape[i]:dim_i]
        # for dim_i.
        slice_begin = tf.shape(x) + pad - shape

    x = tf.pad(x, paddings, constant_values=pad_val)
    x = tf.slice(x, slice_begin, shape)

    return tf.reshape(x, shape)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from keras_cv.src.backend import config

if config.keras_3():
    from keras.ops import *  # noqa: F403, F401
    from keras.preprocessing.image import smart_resize  # noqa: F403, F401

    from keras_cv.src.backend import keras

    name_scope = keras.name_scope
else:
    try:
        from keras.src.ops import *  # noqa: F403, F401
        from keras.src.utils.image_utils import smart_resize  # noqa: F403, F401
    # Import error means Keras isn't installed, or is Keras 2.
    except ImportError:
        from keras_core.src.backend import vectorized_map  # noqa: F403, F401
        from keras_core.src.ops import *  # noqa: F403, F401
        from keras_core.src.utils.image_utils import (  # noqa: F403, F401
            smart_resize,
        )
    if config.backend() == "tensorflow":
        from keras_cv.src.backend.tf_ops import *  # noqa: F403, F401



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing


@keras_cv_export("keras_cv.layers.RandomChannelShift")
class RandomChannelShift(BaseImageAugmentationLayer):
    """Randomly shift values for each channel of the input image(s).

    The input images should have values in the `[0-255]` or `[0-1]` range.

    Input shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `channels_last` format.

    Output shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `channels_last` format.

    Args:
        value_range: The range of values the incoming images will have.
            Represented as a two number tuple written [low, high].
            This is typically either `[0, 1]` or `[0, 255]` depending
            on how your preprocessing pipeline is set up.
        factor: A scalar value, or tuple/list of two floating values in
            the range `[0.0, 1.0]`. If `factor` is a single value, it will
            interpret as equivalent to the tuple `(0.0, factor)`. The `factor`
            will sample between its range for every image to augment.
        channels: integer, the number of channels to shift, defaults to 3 which
            corresponds to an RGB shift. In some cases, there may ber more or
            less channels.
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    rgb_shift = keras_cv.layers.RandomChannelShift(value_range=(0, 255),
        factor=0.5)
    augmented_images = rgb_shift(images)
    ```
    """

    def __init__(self, value_range, factor, channels=3, seed=None, **kwargs):
        super().__init__(**kwargs, seed=seed)
        self.seed = seed
        self.value_range = value_range
        self.channels = channels
        self.factor = preprocessing.parse_factor(factor, seed=self.seed)

    def get_random_transformation(
        self, image=None, label=None, bounding_boxes=None, **kwargs
    ):
        shifts = []
        for _ in range(self.channels):
            shifts.append(self._get_shift())
        return shifts

    def _get_shift(self):
        invert = preprocessing.random_inversion(self._random_generator)
        return tf.cast(invert * self.factor() * 0.5, dtype=self.compute_dtype)

    def augment_image(self, image, transformation=None, **kwargs):
        image = preprocessing.transform_value_range(
            image, self.value_range, (0, 1), dtype=self.compute_dtype
        )
        unstack_rgb = tf.unstack(image, axis=-1)

        result = []
        for c_i in range(self.channels):
            result.append(unstack_rgb[c_i] + transformation[c_i])

        result = tf.stack(
            result,
            axis=-1,
        )
        result = tf.clip_by_value(result, 0.0, 1.0)
        image = preprocessing.transform_value_range(
            result, (0, 1), self.value_range, dtype=self.compute_dtype
        )
        return image

    def augment_bounding_boxes(self, bounding_boxes, **kwargs):
        return bounding_boxes

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def augment_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        return segmentation_mask

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "factor": self.factor,
                "channels": self.channels,
                "value_range": self.value_range,
                "seed": self.seed,
            }
        )
        return config



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_backbone import (
    MobileNetV3Backbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.utils.python_utils import classproperty

ALIAS_DOCSTRING = """MobileNetV3Backbone model with {num_layers} layers.

    References:
        - [Searching for MobileNetV3](https://arxiv.org/abs/1905.02244)
        - [Based on the Original keras.applications MobileNetv3](https://github.com/keras-team/keras/blob/master/keras/applications/mobilenet_v3.py)

    For transfer learning use cases, make sure to read the
    [guide to transfer learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        include_rescaling: bool, whether to rescale the inputs. If set to
            True, inputs will be passed through a `Rescaling(scale=1 / 255)`
            layer. Defaults to True.
        input_shape: optional shape tuple, defaults to (None, None, 3).
        input_tensor: optional Keras tensor (i.e., output of `layers.Input()`)
            to use as image input for the model.

    Example:
    ```python
    input_data = tf.ones(shape=(8, 224, 224, 3))

    # Randomly initialized backbone
    model = {name}Backbone()
    output = model(input_data)
    ```
"""  # noqa: E501


@keras_cv_export("keras_cv.models.MobileNetV3SmallBackbone")
class MobileNetV3SmallBackbone(MobileNetV3Backbone):
    def __new__(
        cls,
        include_rescaling=True,
        input_shape=(None, None, 3),
        input_tensor=None,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "include_rescaling": include_rescaling,
                "input_shape": input_shape,
                "input_tensor": input_tensor,
            }
        )
        return MobileNetV3Backbone.from_preset("mobilenet_v3_small", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "mobilenet_v3_small_imagenet": copy.deepcopy(
                backbone_presets["mobilenet_v3_small_imagenet"]
            ),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations."""
        return cls.presets


@keras_cv_export("keras_cv.models.MobileNetV3LargeBackbone")
class MobileNetV3LargeBackbone(MobileNetV3Backbone):
    def __new__(
        cls,
        include_rescaling=True,
        input_shape=(None, None, 3),
        input_tensor=None,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "include_rescaling": include_rescaling,
                "input_shape": input_shape,
                "input_tensor": input_tensor,
            }
        )
        return MobileNetV3Backbone.from_preset("mobilenet_v3_large", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "mobilenet_v3_large_imagenet": copy.deepcopy(
                backbone_presets["mobilenet_v3_large_imagenet"]
            ),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


setattr(
    MobileNetV3LargeBackbone,
    "__doc__",
    ALIAS_DOCSTRING.format(name="MobileNetV3Large", num_layers="28"),
)
setattr(
    MobileNetV3SmallBackbone,
    "__doc__",
    ALIAS_DOCSTRING.format(name="MobileNetV3Small", num_layers="14"),
)
